import numpy as np
from typing import List, Union, Tuple

# import from package
from resistics.calculators.calculator import Calculator
from resistics.utilities.utilsConfig import loadConfig
from resistics.utilities.utilsEvalFreq import getEvaluationFreqSize
from resistics.utilities.utilsPrint import (
    generalPrint,
    warningPrint,
    blockPrint,
    arrayToString,
    errorPrint,
)


class DecimationParams(Calculator):
    """Class to calculate and hold decimation parameters

    Attributes
    ----------
    sampleFreq : float
        Sampling frequency in Hz.
    divFactor : int
        Minimum division factor when working out decimation factors
    numLevels : int
        Number of decimation levels
    freqPerLevel : int
        Number of frequencies per decimation level
    decFactors : np.ndarray
        Incremental decimation factors
    decFrequencies : np.ndarray
        Sampling frequencies at each decimation level
    evalFreq : np.ndarray 
        Array of evaluation frequencies
    evalFreqPerLevel : np.ndarray
        2-D array (numLevels,freqPerLevel) of evaluation frequencies

    Methods
    -------
    __init__(sampleFreq)
        Initialise with sampling frequency
    getSampleFreqLevel(decLevel)
        Get the sample frequency of the level
    getDecFactor(decLevel)
        Get the decimation factor from level 0 to decLevel
    getIncrementalFactor(decLevel)
        Get the decimation factor from decLevel-1 to decLevel
    getEvalFrequenciesPerLevel(decLevel)
        Returns evaluation frequencies for decLevel
    setFrequencyParams(evalFreq, freqPerLevel, maxLevel)
        Set decimation parameters by evaluation frequencies and frequencies per level
    setDecimationParams(numLevels, freqPerLevel)
        Set decimation parameters by number of levels and frequencies per level
    calcFrequencyParams(numLevels, freqPerLevel)
        Calculate decimation parameters based on number of levels and frequencies per level
    calcDecimationParams(evalFreq, maxLevel, freqPerLevel)
        Calculate decimation paratmers based on evaluation frequencies, frequecies per level and max level
    calcNearestFactor(freq)
        Calculate nearest decimation factor
    printList()
        Class status returned as list of strings
    """

    def __init__(self, sampleFreq: float):
        """Initialise decimation parameters with sampling frequency

        Calculates decimation factors and evaluation frequencies based on defaults

        Parameters
        ----------
        sampleFreq : float
            Sampling frequency
        """

        self.sampleFreq = sampleFreq
        config = loadConfig()
        self.divFactor: int = 2
        self.numLevels: int = config["Decimation"]["numlevels"]
        self.freqPerLevel: int = config["Frequencies"]["perlevel"]
        self.decFactors: np.ndarray
        self.decFrequencies: np.ndarray
        self.evalFreq: np.ndarray
        self.evalFreqPerLevel: np.ndarray

        # calculate some initial values decimation parameters based on defaults
        self.calcFrequencyParams(self.numLevels, self.freqPerLevel)

    def getSampleFreqLevel(self, decLevel: int) -> float:
        """Get sampling frequency for decimation level

        Parameters
        ----------
        decLevel : int
            Decimation level

        Returns
        -------
        out : float
            Sample frequency for decimation level decLevel
        """

        self.checkDecimationLevel(decLevel)
        return self.sampleFreq / self.getDecFactor(decLevel)

    def getDecFactor(self, decLevel: int) -> float:
        """Get decimation factor for decimation level

        Returns decimation factor relative to level 0

        Parameters
        ----------
        decLevel : int
            Decimation level

        Returns
        -------
        out : float
            Decimation factor for level decLevel
        """

        self.checkDecimationLevel(decLevel)
        return self.decFactors[decLevel]

    def getIncrementalFactor(self, decLevel: int) -> int:
        """Get decimation factor for decimation level

        Returns decimation factor relative to decLevel - 1

        Parameters
        ----------
        decLevel : int
            Decimation level

        Returns
        -------
        out : int
            Incremental decimation factor from decLevel - 1
        """

        self.checkDecimationLevel(decLevel)
        if decLevel == 0:
            return int(self.decFactors[decLevel])
        else:
            return int(self.decFactors[decLevel] / self.decFactors[decLevel - 1])

    def getEvalFrequenciesForLevel(self, decLevel: int) -> np.ndarray:
        """Get evaluation frequencies for level decLevel

        Parameters
        ----------
        decLevel : int
            Decimation level

        Returns
        -------
        out : np.ndarray
            Array of evaluation frequencies for level decLevel
        """

        self.checkDecimationLevel(decLevel)
        return self.evalFreqPerLevel[decLevel, :]

    def checkDecimationLevel(self, decLevel: int) -> None:
        """Checks user provided decimation levels and quits if out of bounds

        Parameters
        ----------
        decLevel : int
            Decimation level
        """

        if decLevel < 0:
            self.printError("Decimation level must be greater than 0", quitRun=True)
        if decLevel > self.numLevels - 1:
            self.printError(
                "Decimation level must be less than or equal to {}".format(
                    self.numLevels - 1
                ),
                quitRun=True,
            )

    def setFrequencyParams(
        self, evalFreq: List, freqPerLevel: int, maxLevel: int
    ) -> None:
        """Sets frequency parameters and calculates decimation paramters based on those

        Parameters
        ----------
        evalFreq : List
            Evaluation frequencies
        freqPerLevel : int
            Number of evaluation frequencies per level
        maxLevel : int
            Maximum decimation level
        """

        evalFreq[::-1].sort()  # sort in descending order
        self.calcDecimationParams(evalFreq, freqPerLevel, maxLevel)

    def setDecimationParams(self, numLevels: int, freqPerLevel: int) -> None:
        """Sets number of levels and frequencies per levels to calculate decimation parameters

        Parameters
        ----------
        numLevels : int
            Number of decimation levels
        freqPerLevel : int
            Number of evaluation frequencies per level
        """

        self.calcFrequencyParams(numLevels, freqPerLevel)

    def calcFrequencyParams(self, numLevels: int, freqPerLevel: int) -> None:
        """Calculates evaluation frequencies and decimation parameters

        Uses number of levels and frequencies per levels to auto-calculate evaluation frequencies and then the decimation parameters.

        Parameters
        ----------
        numLevels : int
            Number of decimation levels
        freqPerLevel : int
            Number of evaluation frequencies per level
        """

        numFreq = numLevels * freqPerLevel
        evalFreq = getEvaluationFreqSize(self.sampleFreq, numFreq)
        self.calcDecimationParams(evalFreq, numLevels, freqPerLevel)

    def calcDecimationParams(
        self, evalFreq: Union[np.ndarray, List], maxLevel: int, freqPerLevel: int
    ) -> None:
        """Calculate decimation parameters from evaluation frequencies

        Uses evaluation frequencies, number of frequencies per decimation level and max allowable decimation level to calculate decimation parameters.

        Parameters
        ----------
        evalFreq : List, np.ndarray
            Number of decimation levels
        maxLevel : int
            Maximum allowable number of decimation levels
        freqPerLevel :
            Number of frequencies per level
        """

        evalFreq = np.array(evalFreq)
        # calculating decimation parameters from evaluation frequencies
        maxf = self.sampleFreq / 4
        # find the first evaluation frequency less than or equal to maxf
        fHigh = evalFreq[0]
        for ifreq in range(0, evalFreq.size):
            if evalFreq[ifreq] <= maxf:
                fHigh = evalFreq[ifreq]
                break
        iHigh = evalFreq.tolist().index(fHigh)
        evalFreqSub = evalFreq[iHigh:]
        # calculate number of levels
        numLevels = maxLevel
        # check if enough evaluation frequencies
        if len(evalFreqSub) < freqPerLevel * maxLevel:
            # numLevels = int(math.floor(len(evalFreqSub)/freqPerLevel))
            numLevels = int(np.ceil(1.0 * len(evalFreqSub) / freqPerLevel))

        # do another subslice
        evalFreqSub = evalFreqSub[: numLevels * freqPerLevel]

        # now create an array of evalation frequencies per decimation level
        # evalFreqPerLevel = np.ones(shape=(numLevels, freqPerLevel))
        evalFreqPerLevel = np.ones(shape=(numLevels, freqPerLevel)) * -1
        for ilevel in range(0, numLevels):
            for ifreq in range(0, freqPerLevel):
                if ilevel * freqPerLevel + ifreq >= len(evalFreqSub):
                    break
                evalFreqPerLevel[ilevel, ifreq] = evalFreqSub[
                    ilevel * freqPerLevel + ifreq
                ]

        # now calculate decimation factors
        decFactors = np.ones(shape=(numLevels))
        decFrequencies = np.ones(shape=(numLevels))
        for ilevel in range(0, numLevels):
            decFactors[ilevel], decFrequencies[ilevel] = self.calcNearestFactor(
                evalFreqPerLevel[ilevel][0]
            )

        # finally, set all parameters
        self.evalFreq = evalFreqSub
        self.freqPerLevel = freqPerLevel
        self.numLevels = numLevels
        self.evalFreqPerLevel = evalFreqPerLevel
        self.decFactors = decFactors
        self.decFrequencies = decFrequencies

    def calcNearestFactor(self, freq: float):
        """Calculates nearest decimation factor given frequency

        Parameters
        ----------
        freq : float
            Frequency for which to calculate decimation factor

        Returns
        -------
        fac : int
            The decimation factor
        f : float
            Sampling frequency in Hz given the decimation factor (sampleFreq/decimationFactor)
        """

        # want sampling frequency to be 4 times greater than highest freq
        fsMin = freq * 4
        # set to initial sampling frequency
        f = float(self.sampleFreq)
        fac = 1
        while f > fsMin * self.divFactor:
            f = f / self.divFactor
            fac = fac * self.divFactor
        return fac, f

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = []
        textLst.append("Sampling frequency = {:f}".format(self.sampleFreq))
        textLst.append("Number of decimation levels = {:d}".format(self.numLevels))
        # decimation factors
        for il in range(0, self.numLevels):
            textLst.append(
                "Level = {:d}\tsample freq. [Hz] = {:.6f}\tsample rate [s] = {:.6f}\tdec. factor = {:07d}\tinc. factor = {:d}".format(
                    il,
                    self.decFrequencies[il],
                    1.0 / self.decFrequencies[il],
                    int(self.decFactors[il]),
                    self.getIncrementalFactor(il),
                )
            )
        # evaluation frequencies
        textLst.append("Evaluation frequencies [Hz]")
        for il in range(0, self.numLevels):
            freqForLevel = self.getEvalFrequenciesForLevel(il)
            eFreqStr = arrayToString(freqForLevel)
            textLst.append("Level = {:d}: {}".format(il, eFreqStr))
        return textLst



import os
import numpy as np

# import from package
from resistics.ioHandlers.dataWriter import DataWriter


class DataWriterAscii(DataWriter):
    """Write out ascii data files

    This is simply header files and ascii data files. The header file saved is relevant only to this software and needs to be read in using DataReaderInternal.
    The header file means less processing to read the header information

	Methods
	-------
    setExtension()
        Set the data file extension
    writeDataFiles(chans, timeData)
        Write out time series data
	"""

    def setExtension(self) -> None:
        """For subclasses to set their own extension type"""
        
        self.extension = ".ascii"

    def writeDataFiles(self, chans, timeData):
        self.extension = ".ascii"
        for idx, c in enumerate(chans):
            writePath = os.path.join(
                self.getOutPath(), "chan_{:02d}{}".format(idx, self.extension)
            )
            # this could probably be made quicker - numpy savetxt maybe
            dataF = open(writePath, "w")
            size = timeData.data[c].size
            for i in range(0, size):
                dataF.write("{:9f}\n".format(timeData.data[c][i]))
            dataF.close()



import os
from datetime import date, time, datetime, timedelta
from typing import List

# import from package
from resistics.calculators.calculator import Calculator
from resistics.dataObjects.projectData import ProjectData
from resistics.ioHandlers.spectrumReader import SpectrumReader
from resistics.dataObjects.statisticData import StatisticData
from resistics.ioHandlers.statisticIO import StatisticIO
from resistics.dataObjects.maskData import MaskData
from resistics.utilities.utilsPrint import blockPrint, generalPrint, warningPrint


class MaskCalculator(Calculator):
    """Calculate window masking given constraints

    Calculate masks for time windows. The window masker reads in statistics and writes out a file with windows which match the given constraints for each statistic. The masked file will then go into the window WindowSelector, which passes on shared windows to the processor.    

    Attributes
    ----------
    proj : Project
        A Project object
    maskData : MaskData
        A MaskData object
    specDir : str
        The spectra directory

    Methods
    -------
    __init__(proj, maskData)
        Initialise with a Project instance and MaskData instance
    clearMaskWindows()
        Clear calculated masks
    applyConstraints(site)
        Apply masking constraints
    printList()
        Class status returned as list of strings
    """

    def __init__(self, projData: ProjectData, maskData: MaskData, **kwargs) -> None:
        """Initialise with project and maskData
    
        Parameters
        ----------
        proj : ProjectData
            The project to which masking should be applied
        maskData : MaskData
            A MaskData instance 
        """

        self.projData = projData
        # stats is the stat to use
        self.maskData = maskData
        self.sampleFreq = maskData.sampleFreq
        self.specdir: str = "spectra"
        if "specdir" in kwargs:
            self.specdir = kwargs["specdir"]

    # def getMaskData(self):
    #     return self.maskData

    def clearMaskWindows(self) -> None:
        """Clear mask windows in maskData object """

        self.maskData.maskWindows = {}
        self.maskData.resetMaskWindows()

    def applyConstraints(self, site: str) -> None:
        """Apply the masking constraints to find windows to be excluded
    
        WindowsSelector will later exclude the masked windows before processing

        Parameters
        ----------
        site : str
            The site in the project to mask
        """

        # run through all the statfiles listed
        siteData = self.projData.getSiteData(site)
        measurements = siteData.getMeasurements(self.sampleFreq)
        evalFreqs = self.maskData.evalFreq
        # create a single statisticIO instance to do the reading
        statIO = StatisticIO()
        # loop over stats
        for stat in self.maskData.stats:
            # loop over decimation levels
            for iDec in range(0, self.maskData.numLevels):
                # loop over measurement folders
                for meas in measurements:
                    # try and open the file
                    statIO.setDatapath(
                        os.path.join(
                            siteData.getMeasurementStatPath(meas), self.specdir
                        )
                    )
                    statData = statIO.read(stat, iDec)
                    if not statData:
                        # to next measurement
                        self.printWarning(
                            "No statistic data found for {}, measurement {} and decimation level {}".format(
                                site, meas, iDec
                            )
                        )
                        continue 
                    numWindows = statData.numWindows
                    winStats = statData.winStats
                    for iW in range(0, numWindows):
                        # now check each window
                        winVals = statData.getStatLocal(iW)
                        # now loop over evaluation frequencies
                        for eIdx, eFreq in enumerate(evalFreqs[iDec]):
                            constraintCheck = True
                            freqVal = winVals[eIdx]
                            for component in self.maskData.constraints[eFreq][stat]:
                                index = winStats.index(component)
                                componentVal = freqVal[index]
                                test = (
                                    componentVal
                                    > self.maskData.constraints[eFreq][stat][component][
                                        0
                                    ]
                                    and componentVal
                                    < self.maskData.constraints[eFreq][stat][component][
                                        1
                                    ]
                                )
                                if self.maskData.insideOut[eFreq][stat][component]:
                                    test = not test
                                constraintCheck = constraintCheck and test
                            if not constraintCheck:
                                # if the stat fails, add it - this is a list of windows to exclude
                                self.maskData.maskWindows[eFreq].add(
                                    statData.getGlobalIndex(iW)
                                )

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = []
        textLst.append("Sample Frequency [Hz] = {}".format(self.sampleFreq))
        textLst.append("Spectra directory = {}".format(self.specdir))
        textLst.append("Mask Data information")
        textLst = textLst + self.maskData.printList()
        return textLst



import os
import sys
import itertools
import glob
from datetime import datetime
from typing import List, Tuple

# import from package
from resistics.utilities.utilsPrint import generalPrint, warningPrint, errorPrint


def getDataDirectoryFormats() -> List[str]:
    """Get list of data directory formats   

    Returns
    -------
    out : List[str]
        A list of allowable data directory formats
    """

    return ["meas", "run", "phnx"]


def getDirectoryContents(path: str) -> Tuple[List, List]:
    """Get contents of directory

    Includes both files and directories
    
    Parameters
    ----------
    path : str
        Parent directory path

    Returns
    -------
    dirs : list
        List of directories
    files : list
        List of files excluding hidden files
    """

    if not checkDirExistence(path):
        return [], []  # return empty lists
    dirList = os.listdir(path)  # current directory
    dirs = []
    files = []
    for d in dirList:
        if os.path.isdir(os.path.join(path, d)):
            dirs.append(d)
        else:
            files.append(d)
    return dirs, removeHiddenFiles(files)


def getFilesInDirectory(path: str) -> List:
    """Get files in directory

    Excludes hidden files
    
    Parameters
    ----------
    path : str
        Parent directory path

    Returns
    -------
    files : list
        List of files excluding hidden files
    """

    dirs, files = getDirectoryContents(path)
    return files


def getDirsInDirectory(path: str):
    """Get subdirectories in directory

    Excludes hidden files
    
    Parameters
    ----------
    path : str
        Parent directory path

    Returns
    -------
    dirs : list
        List of subdirectories
    """

    dirs, files = getDirectoryContents(path)
    return dirs


def removeHiddenFiles(files: List) -> List:
    """Remove hidden files from list of files

    Hidden files are those which begin with a .
    
    Parameters
    ----------
    files : list
        List of files

    Returns
    -------
    files : list
        List of files with hidden files removed
    """

    filesNew = []
    for f in files:
        if f[0] != ".":
            filesNew.append(f)
    return filesNew


def getDataDirsInDirectory(path: str) -> List:
    """Get subdirectories in directory

    This uses known data formats as defined by TODO
    
    Parameters
    ----------
    path : str
        Parent directory path

    Returns
    -------
    dirs : list
        List of directories containing time data
    """

    dirs = getDirsInDirectory(path)
    dirsData = []
    formats = getDataDirectoryFormats()
    for d in dirs:
        for format in formats:
            if format in d:
                dirsData.append(d)
    return dirsData


def checkDirExistence(path: str) -> bool:
    """Check if directory exists

    TODO: Should check that it is actually a directory
    
    Parameters
    ----------
    path : str
        Path to check

    Returns
    -------
    out : bool
        True if directory exists
    """

    if not os.path.exists(path):
        return False
    return True


def makeDir(path: str) -> None:
    """Make directory
    
    Parameters
    ----------
    path : str
        Directory path to make
    """

    os.makedirs(path)


def checkAndMakeDir(path: str) -> None:
    """Check if directory exists and make if not
    
    Parameters
    ----------
    path : str
        Directory path to make
    """

    if not checkDirExistence(path):
        makeDir(path)


def checkFilepath(path: str) -> bool:
    """Check if file exists
    
    TODO: Should check that it is actually a file

    Parameters
    ----------
    path : str
        Filepath to check

    Returns
    -------
    out : bool
        True if file exists
    """

    if not os.path.exists(path):
        generalPrint(
            "utilsio::checkFilepath", "File path {} could not be found.".format(path)
        )
        return False
    return True


def fileFormatSampleFreq(sampleFreq: float) -> str:
    """Provide a consistent way to represent floating numbers in filenames

    Parameters
    ----------
    sampleFreq : float
        The sampling frequency
    
    Returns
    -------
    str
        Float converted for string for the purposes of a filename
    """

    sampleFreqStr: str = "{:.3f}".format(sampleFreq)
    return sampleFreqStr.replace(".", "_")


def lineToKeyAndValue(line: str, delim="=") -> Tuple[str, str]:
    """Helper function to read headers

    Parameters
    ----------
    line : str
        A string representing a header line
    delim : str
        The delimeter separating key and value
    
    Returns
    -------
    key : str
        The header key
    val : str
        The header value
    """

    split = line.split(delim)
    key = split[0].strip()
    val = split[1].strip()
    return key, val



import numpy as np
import numpy.fft as fft
import math


def getFrequencyArray(fs: float, samples: int) -> np.ndarray:
    """Calculate the frequency array for a real fourier transform

    Frequency array goes from 0Hz to Nyquist. Nyquist = fs/2
    
    Parameters
    ----------
    fs : float
        Sampling frequency
    samples : int
        Number of samples    

    Returns
    -------
    frequencies : np.ndarray
        Array of rfft frequencies
    """

    nyquist: float = 1.0 * fs / 2.0
    return np.linspace(0, nyquist, samples)


def forwardFFT(data: np.ndarray, norm: bool = True):
    """Forward real fourier transform
    
    Parameters
    ----------
    data : np.ndarray
        Time array to be transformed   
    norm : bool, optional
        Normalisation option

    Returns
    -------
    fourierData : np.ndarray
        Fourier transformed data
    """

    if not norm:
        return fft.rfft(data, axis=0)
    return fft.rfft(data, norm="ortho", axis=0)


def inverseFFT(data: np.ndarray, length: int, norm: bool = True):
    """Inverse real fourier transform
    
    Parameters
    ----------
    data : np.ndarray
        Time array to be transformed
    length : int
        Length of output time data (to remove padding)
    norm : bool, optional
        Normalisation option

    Returns
    -------
    timeData : np.ndarray
        Inverse fourier transformed data
    """

    if not norm:
        return fft.irfft(data, n=length)
    return fft.irfft(data, n=length, norm="ortho")


def padNextPower2(size: int) -> int:
    """Calculate the amount of padding to next power of 2
    
    Parameters
    ----------
    size : float
        Size of array to be padded   

    Returns
    -------
    padSize : int
        Amout of extra padding required to increase to next power of 2
    """

    next2Power = math.ceil(math.log(size, 2))
    next2Size = math.pow(2, int(next2Power))
    return int(next2Size) - size



import inspect
from typing import List, Dict, Any

# import from package
from resistics.utilities.utilsPrint import generalPrint


def parseKeywords(
    default: Dict[str, Any], keywords: Dict[str, Any], printkw: bool = True
):
    """General print to terminal

    Parameters
    ----------
    default : Dict[str, Any]
        Dictionary of default parameters
    keywords : Dict[str, Any]
        Dictionary of optional keywords
    printkw : bool
        Print out the keywords
    """

    for w in default:
        if w in keywords:
            default[w] = keywords[w]
    if printkw:
        generalPrint(
            "{}::utilsCheck::parseKeywords".format(inspect.stack()[1][3]), str(default)
        )
    return default


def elecChannelsList() -> List:
    """List of acceptable electric channels

    Returns
    -------
    out : List
        List of acceptable electric channels 
    """

    return ["Ex", "Ey"]


def isElectric(chan: str) -> bool:
    """Check if channel is electric
    
    Parameters
    ----------
    chan : str
        Channel name

    Returns
    -------
    out : bool
        True if channel is electric 
    """

    if chan in elecChannelsList():
        return True
    return False


def magChannelsList() -> List:
    """List of acceptable magnetic channels

    Returns
    -------
    out : List
        List of acceptable magnetic channels 
    """

    return ["Hx", "Hy", "Hz", "Bx", "By", "Bz"]


def isMagnetic(chan: str) -> bool:
    """Check if channel is magnetic

    Parameters
    ----------
    chan : str
        Channel name

    Returns
    -------
    out : bool
        True if channel is magnetic 
    """

    if chan in magChannelsList():
        return True
    return False


def consistentChans(chan: str) -> str:
    """Convert channels to ensure consistency

    Parameters
    ----------
    chan : str
        Channel name

    Returns
    -------
    out : str
        Converted channel name 
    """

    standardChans = ["Hx", "Hy", "Hz", "Ex", "Ey"]
    if chan in standardChans:
        return chan
    if chan == "Bx":
        return "Hx"
    if chan == "By":
        return "Hy"
    if chan == "Bz":
        return "Hz"
    # otherwise return chan
    return chan



import sys
import os
import numpy as np
from datetime import datetime
import matplotlib.pyplot as plt
from typing import Union, List, Dict

# import from package
from resistics.dataObjects.projectData import ProjectData
from resistics.dataObjects.siteData import SiteData
from resistics.dataObjects.transferFunctionData import TransferFunctionData
from resistics.calculators.decimator import Decimator
from resistics.calculators.windowSelector import WindowSelector
from resistics.calculators.processorSingleSite import ProcessorSingleSite
from resistics.calculators.processorRemoteReference import ProcessorRemoteReference
from resistics.ioHandlers.spectrumReader import SpectrumReader
from resistics.ioHandlers.transferFunctionReader import TransferFunctionReader
from resistics.ioHandlers.transferFunctionWriter import TransferFunctionWriter
from resistics.utilities.utilsChecks import parseKeywords, isElectric, isMagnetic
from resistics.utilities.utilsIO import checkFilepath, fileFormatSampleFreq
from resistics.utilities.utilsPrint import arrayToString
from resistics.utilities.utilsPlotter import (
    savePlot,
    plotOptionsTransferFunction,
    getTransferFunctionFigSize,
    plotOptionsTipper,
    transferFunctionColours,
)
from resistics.utilities.utilsProject import (
    projectText,
    projectWarning,
    projectError,
    getDecimationParameters,
    getWindowParameters,
    getWindowSelector,
    getSingleSiteProcessor,
    getRemoteReferenceProcessor,
)


def getTransferFunctionData(
    projData: ProjectData, site: str, sampleFreq: float, **kwargs
) -> TransferFunctionData:
    """Get transfer function data

    Parameters
    ----------
    projData : projecData
        The project data
    site : str
        Site to get the transfer functiond data for
    sampleFreq : int, float
        The sampling frequency for which to get the transfer function data
    specdir : str, optional
        The spectra directories used
    postpend : str, optional
        The postpend on the transfer function files
    """

    options: Dict = dict()
    options["specdir"]: str = projData.config.configParams["Spectra"]["specdir"]
    options["postpend"]: str = ""
    options = parseKeywords(options, kwargs)

    # deal with the postpend
    if options["postpend"] != "":
        postpend = "_{}".format(options["postpend"])
    else:
        postpend = options["postpend"]

    siteData = projData.getSiteData(site)
    sampleFreqStr = fileFormatSampleFreq(sampleFreq)
    path = os.path.join(
        siteData.transFuncPath,
        "{:s}".format(sampleFreqStr),
        "{}_fs{:s}_{}{}".format(site, sampleFreqStr, options["specdir"], postpend),
    )
    # check path
    if not checkFilepath(path):
        projectWarning("No transfer function file with name {}".format(path))
        return False

    projectText(
        "Reading transfer function for site {}, sample frequency {}, file {}".format(
            site, sampleFreq, path
        )
    )

    tfReader = TransferFunctionReader(path)
    tfReader.printInfo()
    return tfReader.tfData


def processProject(projData: ProjectData, **kwargs) -> None:
    """Process a project

    Parameters
    ----------
    projData : ProjectData
        The project data instance for the project    
    sites : List[str], optional
        List of sites 
    sampleFreqs : List[float], optional
        List of sample frequencies to process
    specdir : str, optional
        The spectra directories to use
    inchans : List[str], optional
        Channels to use as the input of the linear system
    inputsite : str, optional
        Site from which to take the input channels. The default is to use input and output channels from the same site        
    outchans : List[str], optional
        Channels to use as the output of the linear system
    remotesite : str, optional
        The site to use as the remote site
    remotechans : List[str], optional
        Channels to use from the remote reference site
    crosschannels : List[str], optional
        List of channels to use for cross powers
    masks : Dict, optional
        Masks dictionary for passing mask data    
    datetimes : List, optional
        List of datetime constraints, each one as a dictionary            
    postpend : str, optional
        String to postpend to the transfer function output
    """

    options: Dict = dict()
    options["sites"]: List[str] = projData.getSites()
    options["sampleFreqs"]: List[float] = projData.getSampleFreqs()
    options["specdir"]: str = projData.config.configParams["Spectra"]["specdir"]
    options["inchans"]: List[str] = ["Hx", "Hy"]
    options["inputsite"]: str = ""
    options["outchans"]: List[str] = ["Ex", "Ey"]
    options["remotesite"]: str = ""
    options["remotechans"]: List[str] = options["inchans"]
    options["crosschannels"]: List[str] = []
    options["masks"]: Dict = {}
    options["datetimes"]: List = []
    options["postpend"]: str = ""
    options = parseKeywords(options, kwargs)

    for site in options["sites"]:
        siteData = projData.getSiteData(site)
        siteFreqs = siteData.getSampleFreqs()
        for sampleFreq in siteFreqs:
            # check if not included
            if sampleFreq not in options["sampleFreqs"]:
                continue
            processSite(projData, site, sampleFreq, **options)


def processSite(
    projData: ProjectData, site: str, sampleFreq: Union[int, float], **kwargs
):
    """Process a single sampling frequency for a site

    The site passed is assumed to be the output site (the output channels will come from this site). If channels from a different site are desired to be used as the input channels, this can be done by specifying the optional inputsite argument.

    .. todo:: Give a few different examples here

    Parameters
    ----------
    projData : ProjectData
        The project data instance for the project
    site : str
        Site to process 
    sampleFreq : float, int
        Sample frequency to process
    specdir : str, optional
        The spectra directories to use
    inchans : List[str], optional
        Channels to use as the input of the linear system
    inputsite : str, optional
        Site from which to take the input channels. The default is to use input and output channels from the same site
    outchans : List[str], optional
        Channels to use as the output of the linear system
    remotesite : str, optional
        The site to use as the remote site
    remotechans : List[str], optional
        Channels to use from the remote reference site
    crosschannels : List[str], optional
        List of channels to use for cross powers
    masks : Dict, optional
        Masks dictionary for passing mask data
    datetimes : List, optional
        List of datetime constraints, each one as a dictionary
    postpend : str, optional
        String to postpend to the transfer function output
    """

    options = {}
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options["inchans"] = ["Hx", "Hy"]
    options["inputsite"] = ""
    options["outchans"] = ["Ex", "Ey"]
    options["remotesite"] = ""
    options["remotechans"] = options["inchans"]
    options["crosschannels"] = []
    options["masks"] = {}
    options["datetimes"] = []
    options["postpend"] = ""
    options = parseKeywords(options, kwargs)
    if options["inputsite"] == "":
        options["inputsite"] = site

    projectText("Processing site {}, sampling frequency {}".format(site, sampleFreq))
    siteData = projData.getSiteData(site)

    # define decimation parameters
    decParams = getDecimationParameters(sampleFreq, projData.config)
    decParams.printInfo()
    winParams = getWindowParameters(decParams, projData.config)
    # window selector
    winSelector = getWindowSelector(projData, decParams, winParams)

    # if two sites are duplicated (e.g. input site and output site), winSelector only uses distinct sites. Hence using site and inputSite is no problem even if they are the same
    processSites = []
    if options["remotesite"]:
        processSites = [site, options["inputsite"], options["remotesite"]]
        winSelector.setSites(processSites)
    else:
        # if no remote site, then single site processing
        processSites = [site, options["inputsite"]]
        winSelector.setSites(processSites)

    # add window masks
    if len(list(options["masks"].keys())) > 0:
        for maskSite in options["masks"]:
            if maskSite not in processSites:
                # there is a site in the masks dictionary which is of no interest
                continue
            if isinstance(options["masks"][maskSite], str):
                # a single mask
                winSelector.addWindowMask(maskSite, options["masks"][maskSite])
                continue
            if all(isinstance(item, str) for item in options["masks"][maskSite]):
                # list of masks for the site
                for mask in options["masks"][maskSite]:
                    winSelector.addWindowMask(maskSite, mask)

    # add datetime constraints
    for dC in options["datetimes"]:
        levels = None
        if "levels" in dC:
            levels = dC["levels"]

        if dC["type"] == "datetime":
                winSelector.addDatetimeConstraint(dC["start"], dC["stop"], levels)
        if dC["type"] == "time":
                winSelector.addTimeConstraint(dC["start"], dC["stop"], levels)
        if dC["type"] == "date":
                winSelector.addDateConstraint(dC["date"], levels)

    # calculate the shared windows and print info
    winSelector.calcSharedWindows()
    winSelector.printInfo()
    winSelector.printDatetimeConstraints()
    winSelector.printWindowMasks()
    winSelector.printSharedWindows()
    winSelector.printWindowsForFrequency()

    # now have the windows, pass the winSelector to processors
    outPath = siteData.transFuncPath
    if options["remotesite"]:
        projectText(
            "Remote reference processing with sites: in = {}, out = {}, reference = {}".format(
                options["inputsite"], site, options["remotesite"]
            )
        )
        processor = getRemoteReferenceProcessor(winSelector, outPath, projData.config)
        processor.setRemote(options["remotesite"], options["remotechans"])
    else:
        projectText(
            "Single site processing with sites: in = {}, out = {}".format(
                options["inputsite"], site
            )
        )
        processor = getSingleSiteProcessor(winSelector, outPath, projData.config)

    # add the input and output site
    processor.setInput(options["inputsite"], options["inchans"])
    processor.setOutput(site, options["outchans"])
    if len(options["crosschannels"]) > 0:
        processor.crossChannels = options["crosschannels"]
    processor.postpend = options["postpend"]
    processor.printInfo()
    processor.process()


def viewImpedance(projData: ProjectData, **kwargs) -> None:
    """View impedance tensor data

    Parameters
    ----------
    projData : projecData
        The project data
    sites : List[str], optional
        List of sites to plot transfer functions for
    sampleFreqs : List[float], optional 
        List of samples frequencies for which to plot transfer functions
    polarisations : List[str], optional 
        A list of polarisations to plot. For example, ["ExHx", "ExHy", "EyHx", "EyHy"]
    specdir : str, optional
        The spectra directories used
    postpend : str, optional
        The postpend on the transfer function files
    oneplot : bool, optional
        Plot the polarisation on a single plot
    show : bool, optional
        Show the spectra plot
    save : bool, optional
        Save the plot to the images directory
    plotoptions : Dict
        A dictionary of plot options. For example, set the resistivity y limits using res_ylim, set the phase y limits using phase_ylim and set the xlimits using xlim
    """

    options = {}
    options["sites"] = projData.getSites()
    options["sampleFreqs"] = projData.getSampleFreqs()
    options["polarisations"] = ["ExHx", "ExHy", "EyHx", "EyHy"]
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options["postpend"] = ""
    options["oneplot"] = True
    options["save"] = False
    options["show"] = True
    options["plotoptions"] = plotOptionsTransferFunction()
    options = parseKeywords(options, kwargs)

    # loop over sites
    for site in options["sites"]:
        siteData = projData.getSiteData(site)
        sampleFreqs = set(siteData.getSampleFreqs())
        # find the intersection with the options["freqs"]
        sampleFreqs = sampleFreqs.intersection(options["sampleFreqs"])
        sampleFreqs = sorted(list(sampleFreqs))
        print(sampleFreqs)

        # if prepend is a string, then make it a list
        if isinstance(options["postpend"], str):
            options["postpend"] = [options["postpend"]]

        plotfonts = options["plotoptions"]["plotfonts"]
        # now loop over the postpend options
        for pp in options["postpend"]:
            # add an underscore if not empty
            postpend = "_{}".format(pp) if pp != "" else pp

            if options["plotoptions"]["figsize"] is None:
                figsize = getTransferFunctionFigSize(
                    options["oneplot"], len(options["polarisations"])
                )
            else:
                figsize = options["plotoptions"]["figsize"]
            fig = plt.figure(figsize=figsize)
            mks = ["o", "*", "d", "^", "h"]
            lstyles = ["solid", "dashed", "dashdot", "dotted"]
            colours = transferFunctionColours()

            # loop over sampling frequencies
            includedFreqs = []
            for idx, sampleFreq in enumerate(sampleFreqs):

                tfData = getTransferFunctionData(
                    projData, site, sampleFreq, specdir=options["specdir"], postpend=pp
                )
                if not tfData:
                    continue

                includedFreqs.append(sampleFreq)
                projectText(
                    "Plotting transfer function for site {}, sample frequency {}".format(
                        site, sampleFreq
                    )
                )

                # plot
                mk = mks[idx % len(mks)]
                ls = lstyles[idx % len(lstyles)]
                tfData.viewImpedance(
                    fig=fig,
                    polarisations=options["polarisations"],
                    mk=mk,
                    ls=ls,
                    colours=colours,
                    oneplot=options["oneplot"],
                    res_ylim=options["plotoptions"]["res_ylim"],
                    phase_ylim=options["plotoptions"]["phase_ylim"],
                    xlim=options["plotoptions"]["xlim"],
                    label="{}".format(sampleFreq),
                    plotfonts=options["plotoptions"]["plotfonts"],
                )

            # check if any files found
            if len(includedFreqs) == 0:
                continue

            # sup title
            sub = "Site {}: {}".format(site, options["specdir"] + postpend)
            sub = "{}\nfs = {}".format(sub, arrayToString(includedFreqs, decimals=3))
            st = fig.suptitle(sub, fontsize=plotfonts["suptitle"])
            st.set_y(0.99)
            fig.tight_layout()
            fig.subplots_adjust(top=0.92)

            if options["save"]:
                impath = projData.imagePath
                filename = "transFunction_{}_{}{}".format(
                    site, options["specdir"], postpend
                )
                savename = savePlot(impath, filename, fig)
                projectText("Image saved to file {}".format(savename))

        if not options["show"]:
            plt.close("all")
        else:
            plt.show(block=options["plotoptions"]["block"])


def viewTipper(projData: ProjectData, **kwargs) -> None:
    """View transfer function data

    Parameters
    ----------
    projData : projecData
        The project data
    sites : List[str], optional
        List of sites to plot transfer functions for
    sampleFreqs : List[float], optional 
        List of samples frequencies for which to plot transfer functions
    specdir : str, optional
        The spectra directories used
    postpend : str, optional
        The postpend on the transfer function files
    cols : bool, optional
        Boolean flag, True to arrange tipper plot as 1 row with 3 columns
    show : bool, optional
        Show the spectra plot
    save : bool, optional
        Save the plot to the images directory
    plotoptions : Dict
        A dictionary of plot options. For example, set the resistivity y limits using res_ylim, set the phase y limits using phase_ylim and set the xlimits using xlim
    """

    options = {}
    options["sites"] = projData.getSites()
    options["sampleFreqs"] = projData.getSampleFreqs()
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options["postpend"] = ""
    options["cols"] = True
    options["save"] = False
    options["show"] = True
    options["plotoptions"] = plotOptionsTipper()
    options = parseKeywords(options, kwargs)

    # loop over sites
    for site in options["sites"]:
        siteData = projData.getSiteData(site)
        sampleFreqs = set(siteData.getSampleFreqs())
        # find the intersection with the options["freqs"]
        sampleFreqs = sampleFreqs.intersection(options["sampleFreqs"])
        sampleFreqs = sorted(list(sampleFreqs))

        # if prepend is a string, then make it a list
        if isinstance(options["postpend"], str):
            options["postpend"] = [options["postpend"]]

        plotfonts = options["plotoptions"]["plotfonts"]
        # now loop over the postpend options
        for pp in options["postpend"]:
            # add an underscore if not empty
            postpend = "_{}".format(pp) if pp != "" else pp

            fig = plt.figure(figsize=options["plotoptions"]["figsize"])
            mks = ["o", "*", "d", "^", "h"]
            lstyles = ["solid", "dashed", "dashdot", "dotted"]

            # loop over sampling frequencies
            includedFreqs = []
            for idx, sampleFreq in enumerate(sampleFreqs):

                tfData = getTransferFunctionData(
                    projData, site, sampleFreq, specdir=options["specdir"], postpend=pp
                )
                if not tfData:
                    continue

                includedFreqs.append(sampleFreq)
                projectText(
                    "Plotting tipper for site {}, sample frequency {}".format(
                        site, sampleFreq
                    )
                )

                mk = mks[idx % len(mks)]
                ls = lstyles[idx % len(lstyles)]
                tfData.viewTipper(
                    fig=fig,
                    rows=options["cols"],
                    mk=mk,
                    ls=ls,
                    label="{}".format(sampleFreq),
                    xlim=options["plotoptions"]["xlim"],
                    length_ylim=options["plotoptions"]["length_ylim"],
                    angle_ylim=options["plotoptions"]["angle_ylim"],
                    plotfonts=options["plotoptions"]["plotfonts"],
                )

            # check if any files found
            if len(includedFreqs) == 0:
                continue

            # sup title
            sub = "Site {} tipper: {}".format(site, options["specdir"] + postpend)
            sub = "{}\nfs = {}".format(sub, arrayToString(includedFreqs, decimals=3))
            st = fig.suptitle(sub, fontsize=plotfonts["suptitle"])
            st.set_y(0.99)
            fig.tight_layout()
            fig.subplots_adjust(top=0.85)

            if options["save"]:
                impath = projData.imagePath
                filename = "tipper_{}_{}{}".format(site, options["specdir"], postpend)
                savename = savePlot(impath, filename, fig)
                projectText("Image saved to file {}".format(savename))

        if not options["show"]:
            plt.close("all")
        else:
            plt.show(block=options["plotoptions"]["block"])




import numpy as np
import scipy.signal as signal
import pyfftw
from typing import List, Dict

# import from package
from resistics.calculators.calculator import Calculator
from resistics.dataObjects.spectrumData import SpectrumData
from resistics.dataObjects.timeData import TimeData
from resistics.utilities.utilsConfig import loadConfig


class SpectrumCalculator(Calculator):
    """Class for calculating spectra from time data windows

    The class requires the pyfftw and FFTW3 libraries, which allow for fast fourier transforms for spectra calculations

    Attributes
    ----------
    sampleFreq : float
        Sample frequency of the data
    numSamples : int
        The number of samples in each window
    window : bool (true)
        Flag for applying a window function before fourier transform
    windowFunc : np.ndarray
        Window function to apply to time data before fourier transform
    dataArray : np.ndarray
        numpy array to copy data into before performing fourier transform
    ffotObj : pyfftw.FFTW
        fourier transform object

    Methods
    -------
    __init__(sampleFreq, winSamples)
        Initialise with time data sampling frequency and the number of samples in a window
    calcFourierTransform(timeData)
        Perform fourier transform for timeData and return specData object
    printList()
        Class status returned as list of strings        
    """

    def __init__(self, sampleFreq: float, winSamples: int) -> None:
        """Initialise 
    
        Parameters
        ----------
        sampleFreq : float
            Sampling frequency of time data
        winSamples : int
            Number of samples in a window 
        """

        self.sampleFreq: float = sampleFreq
        self.numSamples: int = winSamples
        config = loadConfig()
        self.window: bool = config["Spectra"]["applywindow"]
        self.windowFunc: np.ndarray = signal.get_window(
            config["Spectra"]["windowfunc"], winSamples
        )
        # create an pyfftw plan
        self.dataArray: np.ndarray = pyfftw.empty_aligned(
            self.numSamples, dtype="float64"
        )
        self.fftObj: pyfftw.FFTW = pyfftw.builders.rfft(self.dataArray)

    def calcFourierCoeff(self, timeData: TimeData) -> SpectrumData:
        """Fast fourier transform of timeData

        Parameters
        ----------
        timeData : TimeData
            A TimeData object

        Returns
        -------
        specData : SpectrumData
            A SpectrumData object
        """

        fftData: Dict = {}
        for c in timeData.chans:
            # first copy data into dataArray
            self.dataArray[:] = timeData.data[c][:]
            # no need to pad, these are usually multiples of two
            # remove mean and window if set
            # detrend
            self.dataArray[:] = signal.detrend(self.dataArray, type="linear")
            if self.window:
                self.dataArray[:] = self.dataArray * self.windowFunc
            # use pytfftw here
            fftData[c] = np.array(self.fftObj())
        # calculate frequency array
        dataSize = fftData[timeData.chans[0]].size
        return SpectrumData(
            windowSize=self.numSamples,
            dataSize=dataSize,
            sampleFreq=timeData.sampleFreq,
            startTime=timeData.startTime,
            stopTime=timeData.stopTime,
            data=fftData,
            comments=timeData.comments
            + ["Fourier transform performed using a hamming window"],
        )

    def printList(self) -> List:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = ["Sample frequency = {:f} [Hz]".format(self.sampleFreq)]
        return textLst



import os
import glob
import numpy as np
import scipy.interpolate as interp
from typing import Dict, List

# import from package
from resistics.calculators.calculator import Calculator
from resistics.dataObjects.calibrationData import CalibrationData
from resistics.dataObjects.timeData import TimeData
from resistics.ioHandlers.calibrationIO import CalibrationIO
from resistics.utilities.utilsChecks import isMagnetic
from resistics.utilities.utilsConfig import loadConfig
from resistics.utilities.utilsCalibrate import getKnownCalibrationFormats, getCalName
from resistics.utilities.utilsFreq import (
    forwardFFT,
    inverseFFT,
    getFrequencyArray,
    padNextPower2,
)
from resistics.utilities.utilsPrint import generalPrint, warningPrint, blockPrint


class Calibrator(Calculator):
    """Class for calibrating a time dataset

    Attributes
    ----------
    calExt : list
        Accepted calibration extensions
    calFormats : list
        The calibration file formats of the respective extensions
    calDir : str
        Directory path of calibration files
    calFiles : list[str]
        List of calibration files found in calibration directory
    numCalFiles : int
        Number of found calibration files in calibration directory
    useTheoretical : bool
        Flag to use theoretical calibration function

    Methods
    -------
    __init__(calDirectory)
        Initialise with calibration directory
    setCalDir(calDirectory)
        Set the calibration directory path
    findCalFiles()
        Find calibration files in calibration directory
    calibrate(timeData, sensor, serial, chopper)
        Calibrate time data
    getCalFile(sensor, serial, chopper)
        Get index of matching calibration flile in calFiles list
    calibrateChan(data, fs, calData)
        Calibrate an individual channel
    interpolateCalData(calData, f, spline)
        Interpolate calibration data to same frequencies as channel data
    getTheoreticalCalData(sensor)
        Get theoretical calibration data
    printList()
        Class status returned as list of strings
    """

    def __init__(self, calDirectory: str) -> None:
        """Set the calibration directory and initialise

        Calibrator will automatically find calibration files in the provided directory
    
        Parameters
        ----------
        calDirectory : str
            Path of directory containing calibration files
        """

        self.calExt, self.calFormats = getKnownCalibrationFormats()
        self.calFiles: List[str] = []
        self.numCalFiles = 0
        self.calDir: str = calDirectory
        self.findCalFiles()
        # set whether to use theoretical calibration functions
        conf = loadConfig()
        self.extend: bool = conf["Calibration"]["extend"]
        self.useTheoretical: bool = conf["Calibration"]["usetheoretical"]

    def setCalDir(self, calDirectory: str) -> None:
        """Set the calibration directory and find files

        Calibrator will automatically find calibration files in the provided directory
    
        Parameters
        ----------
        calDirectory : str
            Path of directory containing calibration files
        """

        self.calDir = calDirectory
        self.findCalFiles()

    def findCalFiles(self) -> None:
        """Find calibration files in calibration directory"""

        self.calFiles.clear()
        # get unique extensions
        extensions = list(set(self.calExt))
        for calExt in extensions:
            # get all files of format cE
            self.calFiles = self.calFiles + glob.glob(
                os.path.join(self.calDir, "*.{}".format(calExt))
            )
        self.numCalFiles = len(self.calFiles)

    def calibrate(
        self,
        timeData: TimeData,
        sensor: Dict[str, str],
        serial: Dict[str, int],
        chopper: Dict[str, bool],
    ) -> TimeData:
        """Calibrate time data

        For each channel in timeData, searches for a matching calibration file based on sensor type, serial number and chopper. If a calibration file is found, the channel is calibrated using the data in the file. If useTheoretical is False and no file is found, the data is not calibrated

        todo:
        If no calibration file is found and the channel is a magnetic data channel, a theoretical function can be used
    
        Parameters
        ----------
        timeData : TimeData
            TimeData object
        sensor : Dict
            Dictionary of sensor information with channels as the key and sensor as the value (sensor is a string)
        serial :
            Dictionary of serial information with channels as the key and sensor as the value (serial is a number)
        chopper :
            Dictionary of chopper information with channels as the key and sensor as the value (chopper is a bool)

        Returns
        -------
        timeData : TimeData
            Calibration TimeData object
        """

        calIO = CalibrationIO()
        # iterate over data
        for chan in timeData.chans:
            # output some info
            self.printText("Calibrating channel {}".format(chan))
            # try and find calibration file
            calFile, calFormat = self.getCalFile(
                sensor[chan], serial[chan], chopper[chan]
            )
            if calFile == "":
                # no file found
                if self.useTheoretical and isMagnetic(chan):
                    # use theoretical
                    calData = self.getTheoreticalCalData(sensor[chan])
                    timeData.data[chan] = self.calibrateChan(
                        timeData.data[chan], timeData.sampleFreq, calData
                    )
                    timeData.addComment(
                        "Channel {} calibrated with theoretical calibration function".format(
                            chan
                        )
                    )
                    continue
                else:
                    self.printText(
                        "No Calibration data found - channel will not be calibrated"
                    )
                    timeData.addComment("Channel {} not calibrated".format(chan))
                    continue  # nothing to do

            # else file found
            # no need to separately apply static gain, already included in cal data
            calIO.refresh(calFile, calFormat, chopper=chopper[chan], extend=self.extend)
            calData = calIO.read()
            self.printText(
                "Calibration file found for sensor {}, serial number {}, chopper {}: {}".format(
                    sensor[chan], serial[chan], chopper[chan], calFile
                )
            )
            self.printText("Format: {}".format(calFormat))
            self.printText(
                "Static gain correction of {} applied to calibration data".format(
                    calData.staticGain
                )
            )
            # calibrate time data
            timeData.data[chan] = self.calibrateChan(
                timeData.data[chan], timeData.sampleFreq, calData
            )
            timeData.addComment(
                "Channel {} calibrated with calibration data from file {}".format(
                    chan, calFile
                )
            )
        # return calibrated time data
        return timeData

    def getCalFile(self, sensor, serial, chopper) -> int:
        """Get calibration file for a sensor, serial and chopper combination
    
        Parameters
        ----------
        sensor : str
            Channel data
        serial : int
            Sampling frequency Hz
        chopper : bool
            Calibration data

        Returns
        -------
        calFile : str
            The mathing calibration file
        calFormat : str
            The calibration format
        """
        
        if self.numCalFiles == 0:
            # no calibration files to calibrate with
            return "", ""

        for extension, fileformat in zip(self.calExt, self.calFormats):
            # get the name for this format - there could be more than one acceptable name
            calNames = getCalName(fileformat, extension, sensor, serial, chopper)
            if calNames is None:
                continue
            # if a string rather than a list
            if isinstance(calNames, str):
                calNames = [calNames]
            # search to find a calibration file with that name and take the first encountered
            for calName in calNames:    
                for calFile in self.calFiles:
                    if calName in calFile:
                        return calFile, fileformat
        # else return empty strings
        return "", ""

    def calibrateChan(
        self, data: np.ndarray, sampleFreq: float, calData: CalibrationData
    ) -> np.ndarray:
        """Calibrate a channel

        Perform fourier transform of channel data, deconvolve (division) sensor impulse response and inverse fourier transform.
    
        Parameters
        ----------
        data : np.ndarray
            Channel data
        sampleFreq : float
            Sampling frequency Hz
        calData : CalibrationData
            Calibration data

        Returns
        -------
        out : np.ndarray
            Calibrated channel data
        """

        # do the forward transform
        dataSize = data.size
        # pad end of array
        data = np.pad(data, (0, padNextPower2(dataSize)), "constant")
        fftData = forwardFFT(data)
        f = getFrequencyArray(sampleFreq, fftData.size)
        # do calibration in frequency domain
        # interpolate calibration info to frequencies in data
        transferFunc = self.interpolateCalData(calData, f)
        # recall, zero element excluded, so start from 1
        # fft zero element should really be zero because average is removed from data
        fftData[1:] = fftData[1:] / transferFunc
        # return the non padded part of the array
        return inverseFFT(fftData, data.size)[:dataSize]

    def interpolateCalData(self, calData: CalibrationData, f: np.ndarray) -> np.ndarray:
        """Interpolation calibration data on to frequency points
    
        Parameters
        ----------
        calData : CalibrationData
            Calibration data
        f : np.ndarray
            Frequency array where calibration data is defined

        Returns
        -------
        out : np.ndarray
            Calibration data interpolated to the frequency array f
        """

        # linear interpolate
        interpFuncMag = interp.interp1d(calData.freqs, calData.magnitude)
        interpFuncPhase = interp.interp1d(calData.freqs, calData.phase)
        return interpFuncMag(f[1:]) * np.exp(1j * interpFuncPhase(f[1:]))

    def getTheoreticalCalData(self, sensor: str) -> np.ndarray:
        """Get theoretical calibration data for magnetic channels
    
        Parameters
        ----------
        sensor : str
            Sensor type

        Returns
        -------
        out : np.ndarray
            Theoretical calibration information
        """

        # should use the sensor to figure out what calibration func
        if "mfs06" in sensor or "MFS06" in sensor:
            return CalibrationIO().unitCalibration()

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = []
        textLst.append("Known extensions and calibration formats")
        textLst.append("Extensions:")
        textLst.append(", ".join(self.calExt))
        textLst.append("Associated formats")
        textLst.append(", ".join(self.calFormats))
        # now print the actual calibration files
        textLst.append("{} calibration files found:".format(self.numCalFiles))
        if self.numCalFiles == 0:
            textLst.append("\t\tNo calibration files found")
        else:
            for calFile in self.calFiles:
                textLst.append("\t\t{}".format(calFile))
        return textLst



from copy import deepcopy
from datetime import datetime, timedelta
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
from matplotlib.dates import (
    DateFormatter,
    DayLocator,
    AutoDateLocator,
    AutoDateFormatter,
)
from typing import List, Dict, Tuple, Union, Any

# import from package
from resistics.dataObjects.dataObject import DataObject
from resistics.utilities.utilsPrint import arrayToString
from resistics.utilities.utilsWindow import gArray2datetime
from resistics.utilities.utilsPlotter import (
    getPlotRowsAndCols,
    getViewFonts,
    colorbar2dOther,
    colorbar2dTime,
)


class StatisticData(DataObject):
    """Class for holding information about a statistic

    Attributes
    ----------
    statName : str
        The name of the statistic
    refTime : datetime.datetime
        The reference time of the project
    sampleFreq : float
        The sampling frequency of the data
    winSize : int
        Window size in samples
    winOlap : int
        Window overlap in samples
    numWindows : int
        Number of windows
    winStats : List[str]
        Values calculated for the statistic
    numStatsPerWindow : int
        The number of statistics per window (length of winStats)
    evalFreq : List, np.ndarray
        The evaluation frequencies for the statistic data
    freq2index : Dict
        Mapping from evaluation frequency to index
    globalIndices : List, np.ndarray
        Array of global indices. Allows local -> global conversion
    global2localMap : Dict
        Dictionary which maps global indices to local. Allows global -> local conversion
    comments : List[str]
        Statistic comments
    dtype : str (default "float")
        The data type of the statistic values
    maxcols : int (default 4)
        The number of columns in plots
    globalDatesStart : List, np.ndarray
        Global window start datetimes
    globalDatesStop : List, np.ndarray
        Global window stop datetimes

    Methods
    -------
    __init__(sampleFreq, numLevels, evalFreq, kwargs)
        Initialise maskData
    getStats(includewindows, maskwindows)
        Get the statistics array and choose to either include some windows or mask some windows
    getMaskedWindows(maskwindows)
        Return the local indices to use given a set of global indices to exclude
    getStatLocal(localIndex)
        Get statistic values for a local window index
    getStatGlobal(globalIndex)
        Get statistic values for a global window index
    getGlobalIndex(localIndex)
        Get the global window index for a local window index
    getGlobalDates()
        Get the global index dates for all the windows
    setStatParams(numWindows, winStats, evalFreq)
        Set the statistic parameters and prepare the data arrays
    addStat(localIndex, globalIndex, stat)
        Add a statistic, providing the local index, the corresponding global index and the statististic values
    getComments()
        Get a deepcopy of the comments    
    addComment(comment)
        Add a comment to the comments
    view(eFreqI, **kwargs) 
        View statistic values for a evaluation frequency (time on x axis, statistic values on y axis)
    histogram(eFreqI, **kwargs)
        View a histogram of the statistic values for each statistic component for a single evaluation frequency
    crossplot(eFreqI, **kwargs)   
        View a crossplot of the data
    addColourbar(plot, cax, title, plotfonts)
        Add a colourbar to a plot    
    addColourbarDates(plot, cax, title, plotfonts)
        Add a coloubar the represents dates to the plot   
    dateTicks(gIndices, dates, timeNum)
        Deal with the dateTicks of the plot
    calcColourData(plotData, val, eFreqI, keywords)
        Calculate colours 
    getRowsCols(maxcols, numStats)
        Get the number of rows and columns for a plot
    printList()
        Class status returned as list of strings   
    """

    def __init__(
        self,
        statName: str,
        refTime: datetime,
        sampleFreq: float,
        winSize: int,
        winOlap: int,
        **kwargs
    ):
        """Initialise statistic data 
    
        Parameters
        ----------
        statName : str
            Name of the statistic
        refTime : datetime.datetime
            Reference time for the project
        sampleFreq : float
            Sample frequency of the data in Hz 
        winSize : int
            The window size in samples
        winOlap : int
            The window overlap in samples
        kwargs : Dict
            Two optional arguments: "name" for statistic name and "stats" for the statistics to use
        """

        self.statName: str = statName
        self.refTime = refTime
        if isinstance(self.refTime, str):
            self.refTime = datetime.strptime(self.refTime, "%Y-%m-%d %H:%M:%S.%f")
        self.sampleFreq: float = sampleFreq
        self.winSize: int = winSize
        self.winOlap: int = winOlap
        # defaults from keywords
        self.numWindows: int = 0
        self.winStats: List[str] = []
        self.numStatsPerWindow: int = 0
        self.evalFreq: Union[List, np.ndarray] = []
        self.freq2index: Dict = {}
        self.globalIndices: Union[List, np.ndarray] = []
        self.global2localMap: Dict = {}
        self.comments: List[str] = []
        self.dtype: str = "float"
        # global dates
        self.globalDatesStarts: Union[List, np.ndarray] = []
        self.globalDatesStops: Union[List, np.ndarray] = []
        # plot params
        self.maxcols: int = 4
        # parse keywords
        self.initialiseFromKeywords(kwargs)

    def initialiseFromKeywords(self, keywords):
        if "numWindows" in keywords:
            self.numWindows = keywords["numWindows"]
        if "winStats" in keywords:
            self.winStats = keywords["winStats"]
        self.numStatsPerWindow = len(self.winStats)
        if "evalFreq" in keywords:
            self.evalFreq = keywords["evalFreq"]
        for idx, eFreq in enumerate(self.evalFreq):
            self.freq2index[eFreq] = idx
        if "stats" in keywords:
            self.stats = keywords["stats"]
        if "globalIndices" in keywords:
            self.globalIndices = keywords["globalIndices"]
            # make the global 2 local map
            for ii in np.arange(0, len(self.globalIndices)):
                self.global2localMap[self.globalIndices[ii]] = ii
        if "comment" in keywords:
            self.comments = keywords["comment"]
        if "dtype" in keywords:
            self.dtype = keywords["dtype"]

    def getStats(
        self,
        includewindows: Union[np.ndarray, List] = [],
        maskwindows: Union[np.ndarray, List] = [],
    ) -> np.ndarray:
        """Get statistics when either selecting a set of indices or wanting to mask a set of indices

        Parameters
        ----------
        includewindows : List, np.ndarray, optional
            Windows to get
        maskwindows : List, np.ndarray, optional
            Windows to exclude
        
        Returns
        -------
        np.ndarray
            Statistics limited by the window selection options
        """

        if len(includewindows) > 0:
            return self.stats[includewindows, :]
        elif len(maskwindows) > 0:
            includewindows = self.getMaskedIndices(maskwindows)
            # if this is false, then want to return all the statistics
            if includewindows:
                return self.stats[includewindows, :]
        return self.stats

    def getMaskedIndices(
        self, maskWindows: Union[List, np.ndarray]
    ) -> Union[List, bool]:
        """Given a list of global windows to mask, returns the set of indices to include in the plot

        Parameters
        ----------
        maskWindows : List, np.ndarray
            Global indices of windows to mask
        
        Returns
        -------
        List
            List of indices to include
        """

        includeWindows = set(self.globalIndices) - set(maskWindows)
        if len(includeWindows) == self.numWindows:
            # nothing to mask
            return False

        indices = []
        for win in includeWindows:
            indices.append(self.global2localMap[win])
        return indices

    def getStatLocal(self, localIndex: int) -> np.ndarray:
        """Get statistics for a local window index

        Parameters
        ----------
        localIndex : int
            Local index of the window

        Returns
        -------
        out : np.ndarray
            Statistics for the local window
        """

        return self.stats[localIndex, :]

    def getStatGlobal(self, globalIndex: int) -> np.ndarray:
        """Get statistics for a local window index

        Parameters
        ----------
        globalIndex : int
            Global index of the window

        Returns
        -------
        out : np.ndarray
            Statistics for the local window
        """

        if globalIndex not in self.global2localMap:
            self.printError(
                "There are two statistics with the same global window index"
            )
            self.printError("This should never happen")
            self.printError("Exiting", quitRun=True)
        localIndex = self.global2localMap[globalIndex]
        return self.stats[localIndex]

    def getGlobalIndex(self, localIndex):
        """Get global index for local index

        Parameters
        ----------
        localIndex : int
            Local index of the window

        Returns
        -------
        globalIndex : int
            Global index for local index
        """

        return self.globalIndices[localIndex]

    def getGlobalDates(self):
        """Get the global start dates and end dates for each window
        
        Returns
        -------
        np.ndarray
            The global window start times
        """

        # want to get the start date for each window here
        if len(self.globalDatesStarts) != self.numWindows:
            self.globalDatesStarts, self.globalDatesStops = gArray2datetime(
                self.globalIndices,
                self.refTime,
                self.sampleFreq,
                self.winSize,
                self.winOlap,
            )
        return self.globalDatesStarts

    def setStatParams(
        self, numWindows: int, winStats: List[str], evalFreq, dtype: Any = None
    ):
        """Set the statistic parameters

        Parameters
        ----------
        numWindows : int
            Number of windows for which the statistic has been or will be calculated
        winStats : int
            The parameters in the statistic
        evalFreq : List
            A list of evaluation frequencies
        dtype : str, optional
            The datatype of the statistic values
        """

        self.numWindows = numWindows
        # details about the statistics
        self.winStats = winStats
        self.numStatsPerWindow = len(winStats)
        # details about the evaluation frequencies
        self.evalFreq = evalFreq
        for idx, eFreq in enumerate(evalFreq):
            self.freq2index[eFreq] = idx
        # data type
        if not (dtype is None):
            self.dtype = dtype
        self.stats = np.empty(
            shape=(self.numWindows, self.evalFreq.size, self.numStatsPerWindow),
            dtype=self.dtype,
        )
        # an array to hold global indices and a dictionary to map them back to local indices
        self.globalIndices = np.empty(shape=(self.numWindows), dtype=int)
        self.global2localMap = {}

    def addStat(self, localIndex: int, globalIndex: int, stat) -> None:
        """Add statistic values for a window for all evaluation frequencies

        Parameters
        ----------
        localIndex : int
            Local index of the window
        globalIndex : int
            Global index of the window
        stat : Dict
            An dictionary of dictionaries. First set of key, values are evaluationFrequencies and dictionaries, second set of key, values are statistic names in winStats and their corresponding value for the window
        """

        # add the data - the data is in the format [eFreq][key][val]
        for idx, eFreq in enumerate(self.evalFreq):
            # do it in this order because sorted
            for ist, st in enumerate(self.winStats):
                self.stats[localIndex, idx, ist] = stat[eFreq][st]
        # then add the global index
        self.globalIndices[localIndex] = globalIndex
        self.global2localMap[globalIndex] = localIndex

    def getComments(self) -> List[str]:
        """Get a deepcopy of the comments
        
        Returns
        -------
        List[str]
            Dataset comments as a list of strings
        """

        return deepcopy(self.comments)

    def addComment(self, comment: str) -> None:
        """Add a new comment

        Parameters
        ----------
        comment : float
            A new comment
        """

        self.comments.append(comment)

    def view(self, eFreqI: int, **kwargs) -> plt.figure:
        """Plot statistics for evaluation frequency index

        Plots a simple scatter of each statistic with datetime on the xaxis (datetime of the window start dates). Number of subplots is equal to numStaStatPerWindow.

        Parameters
        ----------
        eFreqI : int
            Evaluation frequency index
        maskwindows : List, np.ndarray
            Global windows to exclude
        fig : matplotlib.pyplot.figure, optional
            A figure object
        plotfonts : Dict, optional
            A dictionary of plot fonts
        label : str, optional
            Label for the plots
        clim : List, optional
            Limits for colourbar axis
        xlim : List, optional
            Limits for the x axis
        ylim : List, optional
            Limits for the y axis
        colortitle : str, optional
            Title for the colourbar
        legened : bool
            Boolean flag for adding a legend
        
        Returns
        -------
        plt.figure
            Matplotlib figure object        
        """

        # get windows to plot and global dates
        maskWindows = kwargs["maskwindows"] if "maskwindows" in kwargs else []
        plotIndices = self.getMaskedIndices(maskWindows)
        globalDates = self.getGlobalDates()
        eFreq = self.evalFreq[eFreqI]

        # plot params
        nrows, ncols = self.getRowsCols(self.maxcols)
        plotfonts = kwargs["plotfonts"] if "plotfonts" in kwargs else getViewFonts()

        fig: plt.figure = (
            plt.figure(kwargs["fig"].number)
            if "fig" in kwargs
            else plt.figure(figsize=(4 * ncols, 5 * nrows))
        )
        st = fig.suptitle(
            "{} data for evaluation frequency: {}".format(self.statName, eFreq),
            fontsize=plotfonts["suptitle"],
        )
        st.set_y(0.98)

        # plot the data
        for idx, val in enumerate(self.winStats):
            ax = plt.subplot(nrows, ncols, idx + 1)
            plt.title("Value {}".format(val), fontsize=plotfonts["title"])
            label = kwargs["label"] if "label" in kwargs else eFreq

            # limit the data by plotIndices if not False
            plotData = np.squeeze(self.stats[:, eFreqI, idx])
            plotDates = globalDates
            if plotIndices:
                plotData = plotData[plotIndices]
                plotDates = plotDates[plotIndices]

            # the colourdata
            colourbool, colourdata, cmap = self.calcColourData(
                plotData, val, eFreqI, kwargs
            )
            # scatter plot
            if not colourbool:
                scat = plt.scatter(
                    plotDates,
                    plotData,
                    edgecolors="none",
                    marker="o",
                    s=12,
                    label=label,
                )
            else:
                scat = plt.scatter(
                    plotDates,
                    plotData,
                    c=colourdata,
                    edgecolors="none",
                    marker="o",
                    s=12,
                    cmap=cmap,
                    label=label,
                )
                clim = (
                    kwargs["clim"]
                    if ("clim" in kwargs and len(kwargs["clim"]) > 0)
                    else [colourdata.min(), colourdata.max()]
                )
                scat.set_clim(clim)
            # x axis options
            plt.xlabel("Time", fontsize=plotfonts["axisLabel"])
            xlim = (
                kwargs["xlim"]
                if ("xlim" in kwargs and len(kwargs["xlim"]) > 0)
                else [globalDates[0], globalDates[-1]]
            )
            plt.xlim(xlim)
            ax.format_xdata = DateFormatter("%H-%M-%S")
            fig.autofmt_xdate()
            # y axis options
            if "ylim" in kwargs and len(kwargs["ylim"]) > 0:
                plt.ylim(kwargs["ylim"])
            plt.ylabel("Value {}".format(val), fontsize=plotfonts["axisLabel"])
            # set tick sizes
            for label in ax.get_xticklabels() + ax.get_yticklabels():
                label.set_fontsize(plotfonts["axisTicks"])
            plt.grid(True, ls="--")
            # legend
            if "legend" in kwargs and kwargs["legend"]:
                plt.legend(loc=4)

        # show if the figure is not in keywords
        if "fig" not in kwargs:
            if colourbool:
                fig.tight_layout(rect=[0.02, 0.02, 0.85, 0.92])
                cax = fig.add_axes([0.88, 0.10, 0.03, 0.80])
                colourtitle = (
                    kwargs["colortitle"] if "colortitle" in kwargs else "Value"
                )
                self.addColourbar(scat, cax, colourtitle, plotfonts)
            else:
                fig.tight_layout(rect=[0.02, 0.02, 0.98, 0.92]) 
            plt.show()

        return fig

    def histogram(self, eFreqI: int, **kwargs) -> plt.figure:
        """Plot statistics for evaluation frequency index

        Plots a histogram of each statistic with bins on the xaxis and count on the yaxis. Ideal for exploring the distribution of statistic values over the windows.

        Parameters
        ----------
        eFreqI : int
            Evaluation frequency index
        maskwindows : List, np.ndarray
            Global windows to exclude            
        numbins : int
            The number of bins for the histogram data binning
        fig : matplotlib.pyplot.figure, optional
            A figure object
        plotfonts : Dict, optional
            A dictionary of plot fonts
        label : str, optional
            Label for the plots
        xlim : List, optional
            Limits for the x axis
        legened : bool
            Boolean flag for adding a legend
        
        Returns
        -------
        plt.figure
            Matplotlib figure object            
        """

        # deal with maskwindows, which are global indices
        maskWindows = kwargs["maskwindows"] if "maskwindows" in kwargs else []
        plotIndices = self.getMaskedIndices(maskWindows)
        eFreq = self.evalFreq[eFreqI]

        # plot options
        numbins = kwargs["numbins"] if "numbins" in kwargs else 40
        nrows, ncols = self.getRowsCols(self.maxcols)

        plotfonts = kwargs["plotfonts"] if "plotfonts" in kwargs else getViewFonts()
        fig = (
            plt.figure(kwargs["fig"].number)
            if "fig" in kwargs
            else plt.figure(figsize=(4 * ncols, 4 * nrows))
        )
        st = fig.suptitle(
            "{} data for evaluation frequency: {}".format(self.statName, eFreq),
            fontsize=plotfonts["suptitle"],
        )
        st.set_y(0.98)

        # plot the data
        for idx, val in enumerate(self.winStats):
            ax = plt.subplot(nrows, ncols, idx + 1)
            plt.title("Value {}".format(val), fontsize=plotfonts["title"])
            label = kwargs["label"] if "label" in kwargs else eFreq

            # data
            plotData = np.squeeze(self.stats[:, eFreqI, idx])
            if plotIndices:
                plotData = plotData[plotIndices]
            # remove infinities and nans
            plotData = plotData[np.isfinite(plotData)]

            # x axis options
            xlim = (
                kwargs["xlim"]
                if "xlim" in kwargs
                else [np.min(plotData), np.max(plotData)]
            )
            plt.xlim(xlim)
            plt.xlabel("Value", fontsize=plotfonts["axisLabel"])
            # now plot with xlim in mind
            plt.hist(plotData, numbins, range=xlim, facecolor="red", alpha=0.75)
            # y axis options
            plt.ylabel("Count", fontsize=plotfonts["axisLabel"])
            # set tick sizes
            for label in ax.get_xticklabels() + ax.get_yticklabels():
                label.set_fontsize(plotfonts["axisTicks"])
            plt.grid(True, ls="--")
            # legend
            if "legend" in kwargs and kwargs["legend"]:
                plt.legend(loc=4)

        # show if the figure is not in keywords
        if "fig" not in kwargs:
            fig.tight_layout(rect=[0.02, 0.02, 0.98, 0.92])            
            plt.show()

        return fig

    def crossplot(self, eFreqI: int, **kwargs) -> plt.figure:
        """Plots crossplots of statistic components for evaluation frequency index

        Notes
        -----
        By default, the crossplots plotted are
        statistic component 1 vs statistic component 2
        statistic component 3 vs statistic component 4
        etc

        But crossplots can be explicity set by using the crossplots keyword. They should be specified as a list of a list of strings
        e.g.
        crossplots = [[component2, component3], [component1, component4]]

        Parameters
        ----------
        eFreqI : int
            Evaluation frequency index
        maskwindows : List, np.ndarray
            Global windows to exclude              
        crossplots : List[List[str]], optional
            The parameters to crossplot
        fig : matplotlib.pyplot.figure, optional
            A figure object
        plotfonts : Dict, optional
            A dictionary of plot fonts
        label : str, optional
            Label for the plots
        clim : List, optional
            Limits for colourbar axis
        xlim : List, optional
            Limits for the x axis
        ylim : List, optional
            Limits for the y axis
        colortitle : str, optional
            Title for the colourbar
        legened : bool
            Boolean flag for adding a legend
        
        Returns
        -------
        plt.figure
            Matplotlib figure object            
        """

        # deal with maskwindows, which are global indices
        maskWindows = kwargs["maskwindows"] if "maskwindows" in kwargs else []
        plotIndices = self.getMaskedIndices(maskWindows)

        # figure out the crossplots
        if "crossplots" in kwargs:
            crossplots = kwargs["crossplots"]
        else:
            crossplots = list(zip(self.winStats[::2], self.winStats[1::2]))

        # plot parameters
        nrows, ncols = self.getRowsCols(self.maxcols, numStats=len(crossplots))
        eFreq = self.evalFreq[eFreqI]

        plotfonts = kwargs["plotfonts"] if "plotfonts" in kwargs else getViewFonts()
        fig = (
            plt.figure(kwargs["fig"].number)
            if "fig" in kwargs
            else plt.figure(figsize=(4 * ncols, 5 * nrows))
        )
        st = fig.suptitle(
            "{} crossplots for evaluation frequency: {}".format(self.statName, eFreq),
            fontsize=plotfonts["suptitle"],
        )
        st.set_y(0.98)

        # now plot the data
        for idx, cplot in enumerate(crossplots):
            ax = plt.subplot(nrows, ncols, idx + 1)
            plt.title(
                "{} vs. {}".format(cplot[0], cplot[1]), fontsize=plotfonts["title"]
            )
            label = kwargs["label"] if "label" in kwargs else eFreq

            # the colourdata
            colourbool, colourdata, cmap = self.calcColourData(
                self.globalIndices, cplot[0], eFreqI, kwargs
            )

            # get plot data
            plotI1 = self.winStats.index(cplot[0])
            plotData1 = np.squeeze(self.stats[:, eFreqI, plotI1])
            plotI2 = self.winStats.index(cplot[1])
            plotData2 = np.squeeze(self.stats[:, eFreqI, plotI2])
            if plotIndices:
                plotData1 = plotData1[plotIndices]
                plotData2 = plotData2[plotIndices]
                colourdata = colourdata[plotIndices]

            # scatter plot
            scat = plt.scatter(
                plotData1,
                plotData2,
                c=colourdata,
                edgecolors="none",
                marker="o",
                s=12,
                cmap=cmap,
                label=label,
            )
            clim = (
                kwargs["clim"]
                if "clim" in kwargs
                else [colourdata.min(), colourdata.max()]
            )
            scat.set_clim(clim)
            # x axis options
            plt.xlabel("Value {}".format(cplot[0]), fontsize=plotfonts["axisLabel"])
            if "xlim" in kwargs:
                plt.xlim(kwargs["xlim"])
            # y axis options
            plt.ylabel("Value {}".format(cplot[1]), fontsize=plotfonts["axisLabel"])
            if "ylim" in kwargs:
                plt.ylim(kwargs["ylim"])
            # set tick sizes
            for label in ax.get_xticklabels() + ax.get_yticklabels():
                label.set_fontsize(plotfonts["axisTicks"])
            plt.grid(True, ls="--")
            # legend
            if "legend" in kwargs and kwargs["legend"]:
                plt.legend(loc=4)

        # show if the figure is not in keywords
        if "fig" not in kwargs:
            fig.tight_layout(rect=[0.02, 0.02, 0.85, 0.92])
            cax = fig.add_axes([0.88, 0.10, 0.03, 0.80])
            if not colourbool:
                self.addColourbarDates(scat, cax, "Time", plotfonts)
            else:
                colourtitle = (
                    kwargs["colortitle"] if "colortitle" in kwargs else "Value"
                )
                self.addColourbar(scat, cax, colourtitle, plotfonts)
            plt.show()

        return fig

    def addColourbar(self, plot: plt.plot, cax, title: str, plotfonts: Dict) -> None:
        """Add a colourbar to a plot

        Parameters
        ----------
        plot : plt.plot
            Evaluation frequency index
        cax : 
            Colourbar axis
        title : str
            The tital for the colourbar
        plotfonts : Dict
            A dictionary with font types as keys and sizes as values
        """

        plt.colorbar(plot, cax=cax)
        cax.set_title(title, y=1.02, fontsize=plotfonts["title"])

    def addColourbarDates(self, plot: plt.plot, cax, title: str, plotfonts: Dict):
        """Make the colourbar show dates for identifying window times where there is not date axis

        Parameters
        ----------
        plot : plt.plot
            Evaluation frequency index
        cax : 
            Colourbar axis
        title : str
            The tital for the colourbar
        plotfonts : Dict
            A dictionary with font types as keys and sizes as values
        """

        ticks, tickLabels = self.dateTicks(self.globalIndices, self.getGlobalDates(), 5)
        cb = plt.colorbar(plot, cax=cax)
        cb.set_ticks(ticks)
        cb.set_ticklabels(tickLabels)
        cax.set_title(title, y=1.02, fontsize=plotfonts["title"])

    def dateTicks(self, gIndices, dates, timeNum):
        """todo: Write the documentation
        """

        numVals = len(gIndices)
        if timeNum >= numVals:
            timeNum = numVals
        plotIndices = []
        for i in range(0, timeNum):
            plotIndices.append(int(i * numVals * 1.0 / (timeNum - 1)))
        plotIndices[-1] = numVals - 1
        ticks = []
        tickLabels = []
        for i in plotIndices:
            ticks.append(gIndices[i])
            tickLabels.append(dates[i].strftime("%m-%d %H:%M:%S"))
        return ticks, tickLabels

    def calcColourData(self, plotData, val, eFreqI, keywords):
        """todo: Write the documentation
        """

        if "colorstat" in keywords and "colormap" in keywords:
            colourVal = keywords["colormap"][val]
            colourIndex = keywords["colorstat"].winStats.index(colourVal)
            return (
                True,
                np.squeeze(keywords["colorstat"].stats[:, eFreqI, colourIndex]),
                colorbar2dOther(),
            )
        return False, plotData, colorbar2dTime()

    def getRowsCols(self, maxcols: int, numStats: int = 0) -> Tuple[int, int]:
        """Get the numbers of rows and columns for the plots

        Parameters
        ----------
        maxcols : int
            The maximum number of columns
        numStats : int
            The number of statistics to plot (in case this is not all of the winStats)
        """

        if numStats < 1:
            numStats = self.numStatsPerWindow
        return getPlotRowsAndCols(maxcols, numStats)

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst: List[str] = []
        textLst.append("Statistic Name = {}".format(self.statName))
        textLst.append(
            "Reference time = {}".format(self.refTime.strftime("%Y-%m-%d %H:%M:%S.%f"))
        )
        textLst.append("Sample freq [Hz] = {}".format(self.sampleFreq))
        textLst.append("Window size = {}".format(self.winSize))
        textLst.append("Window overlap = {}".format(self.winOlap))
        textLst.append("Number of windows = {}".format(self.numWindows))
        textLst.append("Statistics per Window = {}".format(", ".join(self.winStats)))
        textLst.append(
            "Evalutation frequencies = {}".format(arrayToString(self.evalFreq))
        )
        # comments
        textLst.append("Comments...")
        if len(self.comments) == 0:
            textLst.append("No comments")
        else:
            for comment in self.comments:
                textLst.append("\t{}".format(comment))
        return textLst




import sys
import os
import numpy as np
from datetime import datetime
import matplotlib.pyplot as plt
from matplotlib import cm
from typing import Dict, Union


# import from package
from resistics.dataObjects.projectData import ProjectData
from resistics.dataObjects.statisticData import StatisticData
from resistics.calculators.windowSelector import WindowSelector
from resistics.calculators.statisticCalculator import StatisticCalculator
from resistics.ioHandlers.spectrumReader import SpectrumReader
from resistics.ioHandlers.statisticIO import StatisticIO
from resistics.project.projectMask import getMaskData
from resistics.utilities.utilsIO import fileFormatSampleFreq
from resistics.utilities.utilsPrint import listToString
from resistics.utilities.utilsPlotter import (
    savePlot,
    plotOptionsSpec,
    getPlotRowsAndCols,
)
from resistics.utilities.utilsChecks import parseKeywords
from resistics.utilities.utilsStats import getStatElements
from resistics.utilities.utilsProject import (
    projectText,
    projectWarning,
    projectError,
    getDecimationParameters,
    getWindowParameters,
)


def getStatisticData(
    projData: ProjectData, site: str, meas: str, stat: str, declevel: int = 0, **kwargs
):
    """Get the statistic data for a statistic for a site measurement

    Parameters
    ----------
    projData : ProjectData
        Project instance
    site : str
        The site for which to get the statistic data
    meas : str
        The measurement for which to get the statistic
    stat : str
        The statistic for which to get the measurement
    declevel : int, optional
        The decimation level to read in. Default is 0.
    specdir : str, optional
        The spectra directory

    Returns
    -------
    StatisticData
        A statistic data object
    """

    options = {}
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options = parseKeywords(options, kwargs)

    siteData = projData.getSiteData(site)
    if not siteData:
        projectError("Unable to find site {} in project".format(site), quitRun=True)

    # load the statistic data
    statIO = StatisticIO()
    statIO.setDatapath(
        os.path.join(siteData.getMeasurementStatPath(meas), options["specdir"])
    )
    statData = statIO.read(stat, declevel)
    return statData


def getStatisticDataForSampleFreq(
    projData: ProjectData,
    site: str,
    sampleFreq: float,
    stat: str,
    declevel: int = 0,
    **kwargs
):
    """Get the statistic data (for a particular decimation level) for all measurements in a site with sampling frequency sampleFreq

    Parameters
    ----------
    projData : ProjectData
        Project instance
    site : str
        The site for which to get the statistic data
    sampleFreq : float
        The sampling frequency
    stat : str
        The statistic for which to get the measurement
    declevel : int, optional
        The decimation level to read in. Default is 0.
    specdir : str, optional
        The spectra directory

    Returns
    -------
    StatisticData : Dict
        A statistic data object
    """

    options = {}
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options = parseKeywords(options, kwargs)

    siteData = projData.getSiteData(site)
    if not siteData:
        projectError("Unable to find site {} in project".format(site), quitRun=True)

    # load the statistic data
    statData: Dict[str, StatisticData] = {}
    statIO = StatisticIO()
    measurements = siteData.getMeasurements(sampleFreq)
    for meas in measurements:
        statIO.setDatapath(
            os.path.join(siteData.getMeasurementStatPath(meas), options["specdir"])
        )
        # make sure some data was found
        chk = statIO.read(stat, declevel)
        if chk:
            statData[meas] = statIO.read(stat, declevel)
        else:
            projectWarning(
                "No {} data found for site {} and measurement {}".format(
                    stat, site, meas
                )
            )
    return statData


def calculateStatistics(projData: ProjectData, **kwargs):
    """Calculate statistics for sites
    
    Parameters
    ----------
    projData : ProjectData
        A project data instance
    sites : List[str], optional
        A list of sites to calculate statistics for
    sampleFreqs : List[float], optional
        List of sampling frequencies for which to calculate statistics
    chans : List[str], optional
        List of data channels to use
    specdir : str, optional
        The spectra directory for which to calculate statistics
    stats : List[str], optional
        The statistics to calculate out. Acceptable values are: "absvalEqn" "coherence", "psd", "poldir", "transFunc", "resPhase", "partialcoh". Configuration file values are used by default.
    """

    options = {}
    options["sites"] = projData.getSites()
    options["sampleFreqs"] = projData.getSampleFreqs()
    options["chans"] = []
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options["stats"] = projData.config.configParams["Statistics"]["stats"]
    options = parseKeywords(options, kwargs)

    projectText(
        "Calculating stats: {} for sites: {}".format(
            listToString(options["stats"]), listToString(options["sites"])
        )
    )

    # create the statistic calculator and IO object
    statCalculator = StatisticCalculator()
    statIO = StatisticIO()

    # loop through sites and calculate statistics
    for site in options["sites"]:
        siteData = projData.getSiteData(site)
        measurements = siteData.getMeasurements()

        for meas in measurements:
            sampleFreq = siteData.getMeasurementSampleFreq(meas)
            if sampleFreq not in options["sampleFreqs"]:
                # don't need to calculate statistics for this sampling frequency
                continue

            projectText(
                "Calculating stats for site {}, measurement {}".format(site, meas)
            )

            # decimation parameters
            decParams = getDecimationParameters(sampleFreq, projData.config)
            decParams.printInfo()
            numLevels = decParams.numLevels

            # create the spectrum reader
            specReader = SpectrumReader(
                os.path.join(siteData.getMeasurementSpecPath(meas), options["specdir"])
            )

            # loop through decimation levels
            for iDec in range(0, numLevels):
                # open the spectra file for the current decimation level
                check = specReader.openBinaryForReading("spectra", iDec)
                if not check:
                    # probably because this decimation level not calculated
                    continue
                specReader.printInfo()

                # get windows
                refTime = specReader.getReferenceTime()
                winSize = specReader.getWindowSize()
                winOlap = specReader.getWindowOverlap()
                numWindows = specReader.getNumWindows()
                evalFreq = decParams.getEvalFrequenciesForLevel(iDec)
                sampleFreqDec = specReader.getSampleFreq()
                globalOffset = specReader.getGlobalOffset()
                fArray = specReader.getFrequencyArray()

                statHandlers = {}
                # create the statistic handlers
                for stat in options["stats"]:
                    statElements = getStatElements(stat)
                    statHandlers[stat] = StatisticData(
                        stat, refTime, sampleFreqDec, winSize, winOlap
                    )
                    statHandlers[stat].setStatParams(numWindows, statElements, evalFreq)
                    statHandlers[stat].comments = specReader.getComments()
                    statHandlers[stat].addComment(projData.config.getConfigComment())
                    statHandlers[stat].addComment(
                        "Calculating statistic: {}".format(stat)
                    )
                    statHandlers[stat].addComment(
                        "Statistic components: {}".format(listToString(statElements))
                    )

                # loop over windows and calculate the relevant statistics
                for iW in range(0, numWindows):
                    # get data
                    winData = specReader.readBinaryWindowLocal(iW)
                    globalIndex = iW + globalOffset
                    # give the statistic calculator the spectra
                    statCalculator.setSpectra(fArray, winData, evalFreq)
                    # get the desired statistics
                    for sH in statHandlers:
                        data = statCalculator.getDataForStatName(sH)
                        statHandlers[sH].addStat(iW, globalIndex, data)

                # save statistic
                for sH in statHandlers:
                    statIO.setDatapath(
                        os.path.join(
                            siteData.getMeasurementStatPath(meas), options["specdir"]
                        )
                    )
                    statIO.write(statHandlers[sH], iDec)


def calculateRemoteStatistics(projData: ProjectData, remoteSite: str, **kwargs):
    """Calculate statistics involving a remote reference site

    
    Parameters
    ----------
    projData : ProjectData
        A project data instance
    remoteSite : str
        The name of the site to use as the remote site
    sites : List[str], optional
        A list of sites to calculate statistics for
    sampleFreqs : List[float], optional
        List of sampling frequencies for which to calculate statistics
    chans : List[str], optional
        List of data channels to use
    specdir : str, optional
        The spectra directory for which to calculate statistics
    remotestats : List[str], optional
        The statistics to calculate out. Acceptable statistics are: "coherenceRR", "coherenceRREqn", "absvalRREqn", "transFuncRR", "resPhaseRR". Configuration file values are used by default.
    """

    options = {}
    options["sites"] = projData.getSites()
    options["sampleFreqs"] = projData.getSampleFreqs()
    options["chans"] = []
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options["remotestats"] = projData.config.configParams["Statistics"]["remotestats"]
    options = parseKeywords(options, kwargs)

    projectText(
        "Calculating stats: {} for sites: {} with remote site {}".format(
            listToString(options["remotestats"]),
            listToString(options["sites"]),
            remoteSite,
        )
    )

    # create the statistic calculator and IO object
    statCalculator = StatisticCalculator()
    statIO = StatisticIO()

    # loop over sites
    for site in options["sites"]:
        siteData = projData.getSiteData(site)
        measurements = siteData.getMeasurements()

        for meas in measurements:
            sampleFreq = siteData.getMeasurementSampleFreq(meas)
            if sampleFreq not in options["sampleFreqs"]:
                # don't need to calculate statistics for this sampling frequency
                continue

            projectText(
                "Calculating stats for site {}, measurement {} with reference {}".format(
                    site, meas, remoteSite
                )
            )

            # decimation and window parameters
            decParams = getDecimationParameters(sampleFreq, projData.config)
            decParams.printInfo()
            numLevels = decParams.numLevels
            winParams = getWindowParameters(decParams, projData.config)

            # create the window selector and find the shared windows
            winSelector = WindowSelector(projData, sampleFreq, decParams, winParams)
            winSelector.setSites([site, remoteSite])
            # calc shared windows between site and remote
            winSelector.calcSharedWindows()
            # create the spectrum reader
            specReader = SpectrumReader(
                os.path.join(siteData.getMeasurementSpecPath(meas), options["specdir"])
            )

            # loop through decimation levels
            for iDec in range(0, numLevels):
                # open the spectra file for the current decimation level
                check = specReader.openBinaryForReading("spectra", iDec)
                if not check:
                    # probably because this decimation level not calculated
                    continue
                specReader.printInfo()

                # get a set of the shared windows at this decimation level
                # these are the global indices
                sharedWindows = winSelector.getSharedWindowsLevel(iDec)

                # get other information regarding only this spectra file
                refTime = specReader.getReferenceTime()
                winSize = specReader.getWindowSize()
                winOlap = specReader.getWindowOverlap()
                numWindows = specReader.getNumWindows()
                evalFreq = decParams.getEvalFrequenciesForLevel(iDec)
                sampleFreqDec = specReader.getSampleFreq()
                globalOffset = specReader.getGlobalOffset()
                fArray = specReader.getFrequencyArray()

                # now want to find the size of the intersection between the windows in this file and the shared windows
                sharedWindowsMeas = sharedWindows.intersection(
                    set(np.arange(globalOffset, globalOffset + numWindows))
                )
                sharedWindowsMeas = sorted(list(sharedWindowsMeas))
                numSharedWindows = len(sharedWindowsMeas)

                statHandlers = {}
                # create the statistic handlers
                for stat in options["remotestats"]:
                    statElements = getStatElements(stat)
                    statHandlers[stat] = StatisticData(
                        stat, refTime, sampleFreqDec, winSize, winOlap
                    )
                    # remember, this is with the remote reference, so the number of windows is number of shared windows
                    statHandlers[stat].setStatParams(
                        numSharedWindows, statElements, evalFreq
                    )
                    statHandlers[stat].comments = specReader.getComments()
                    statHandlers[stat].addComment(projData.config.getConfigComment())
                    statHandlers[stat].addComment(
                        "Calculating remote statistic: {}".format(stat)
                    )
                    statHandlers[stat].addComment(
                        "Statistic components: {}".format(listToString(statElements))
                    )

                # loop over the shared windows between the remote station and local station
                for iW, globalWindow in enumerate(sharedWindowsMeas):
                    # get data and set in the statCalculator
                    winData = specReader.readBinaryWindowGlobal(globalWindow)
                    statCalculator.setSpectra(fArray, winData, evalFreq)
                    # for the remote site, use the reader in win selector
                    remoteSF, remoteReader = winSelector.getSpecReaderForWindow(
                        remoteSite, iDec, globalWindow
                    )
                    winDataRR = remoteReader.readBinaryWindowGlobal(globalWindow)
                    statCalculator.addRemoteSpec(winDataRR)

                    for sH in statHandlers:
                        data = statCalculator.getDataForStatName(sH)
                        statHandlers[sH].addStat(iW, globalWindow, data)

                # save statistic
                for sH in statHandlers:
                    statIO.setDatapath(
                        os.path.join(
                            siteData.getMeasurementStatPath(meas), options["specdir"]
                        )
                    )
                    statIO.write(statHandlers[sH], iDec)


def viewStatistic(
    projData: ProjectData, site: str, sampleFreq: Union[int, float], stat: str, **kwargs
) -> Union[plt.figure, None]:
    """View statistic data for a single sampling frequency of a site
    
    Parameters
    ----------
    projData : ProjectData
        A project instance
    site : str
        The site for which to plot statistics
    stat : str
        The statistic to plot
    sampleFreq : float
        The sampling frequency for which to plot statistics
    declevel : int
        The decimation level to plot
    eFreqI : int
        The evaluation frequency index
    specdir : str
        The spectra directory
    maskname : str
        Mask name         
    clim : List, optional
        Limits for colourbar axis
    xlim : List, optional
        Limits for the x axis
    ylim : List, optional
        Limits for the y axis
    colortitle : str, optional
        Title for the colourbar
    show : bool, optional
        Show the spectra plot
    save : bool, optional
        Save the plot to the images directory
    plotoptions : Dict, optional
        Dictionary of plot options    

    Returns
    -------
    matplotlib.pyplot.figure or None
        A matplotlib figure unless the plot is not shown and is saved, in which case None and the figure is closed.
    """

    options = {}
    options["declevel"] = 0
    options["eFreqI"] = 0
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options["maskname"] = ""
    options["clim"] = []
    options["xlim"] = []
    options["ylim"] = []
    options["colortitle"] = ""
    options["show"] = True
    options["save"] = False
    options["plotoptions"] = plotOptionsSpec()
    options = parseKeywords(options, kwargs)

    projectText(
        "Plotting statistic {} for site {} and sampling frequency {}".format(
            stat, site, sampleFreq
        )
    )
    statData = getStatisticDataForSampleFreq(
        projData,
        site,
        sampleFreq,
        stat,
        declevel=options["declevel"],
        specdir=options["specdir"],
    )
    # get the statistics
    statMeas = list(statData.keys())
    # get the evaluation frequency
    eFreq = statData[statMeas[0]].evalFreq[options["eFreqI"]]

    # get the mask data
    maskWindows = []
    if options["maskname"] != "":
        maskData = getMaskData(projData, site, options["maskname"], sampleFreq)
        maskWindows = maskData.getMaskWindowsFreq(
            options["declevel"], options["eFreqI"]
        )

    # setup the figure
    plotfonts = options["plotoptions"]["plotfonts"]
    fig = plt.figure(figsize=options["plotoptions"]["figsize"])

    # get the date limits
    siteData = projData.getSiteData(site)
    if len(options["xlim"]) == 0:
        start = siteData.getMeasurementStart(statMeas[0])
        end = siteData.getMeasurementEnd(statMeas[0])
        for meas in statMeas:
            start = min(start, siteData.getMeasurementStart(meas))
            end = max(end, siteData.getMeasurementEnd(meas))
        options["xlim"] = [start, end]

    # do the plots
    for meas in statMeas:
        statData[meas].view(
            options["eFreqI"],
            fig=fig,
            xlim=options["xlim"],
            ylim=options["ylim"],
            clim=options["clim"],
            label=meas,
            plotfonts=options["plotoptions"]["plotfonts"],
            maskwindows=maskWindows,
        )
    # add a legened
    plt.legend()

    # do the title after all the plots
    st = fig.suptitle(
        "{} values for {}, sampling frequency = {:.2f} Hz, decimation level = {} and evaluation frequency {} Hz".format(
            stat, site, sampleFreq, options["declevel"], eFreq
        ),
        fontsize=plotfonts["suptitle"],
    )

    # plot format, show and save
    fig.tight_layout(rect=[0.02, 0.02, 0.98, 0.92])      
    if options["save"]:
        impath = projData.imagePath
        sampleFreqStr = fileFormatSampleFreq(sampleFreq)
        filename = "stat_{:s}_{:s}_{:s}_dec{:d}_efreq{:d}_{:s}".format(
            stat,
            site,
            sampleFreqStr,
            options["declevel"],
            options["eFreqI"],
            options["specdir"],
        )
        if options["maskname"] != "":
            filename = "{}_{}".format(filename, options["maskname"])
        savename = savePlot(impath, filename, fig)
        projectText("Image saved to file {}".format(savename))
    if options["show"]:
        plt.show(block=options["plotoptions"]["block"])
    if not options["show"] and options["save"]:
        plt.close(fig)
        return None        
    return fig


def viewStatisticHistogram(
    projData: ProjectData, site: str, sampleFreq: float, stat: str, **kwargs
) -> Union[plt.figure, None]:
    """View statistic histograms for a single sampling frequency of a site
    
    Parameters
    ----------
    projData : ProjectData
        A project instance
    site : str
        The site for which to plot statistics
    stat : str
        The statistic to plot
    sampleFreq : float
        The sampling frequency for which to plot statistics
    declevel : int
        The decimation level to plot
    eFreqI : int
        The evaluation frequency index       
    specdir : str
        The spectra directory        
    maskname : str
        Mask name 
    numbins : int
        The number of bins for the histogram data binning
    xlim : List, optional
        Limits for the x axis
    maxcols : int
        The maximum number of columns in the plots
    show : bool, optional
        Show the spectra plot
    save : bool, optional
        Save the plot to the images directory
    plotoptions : Dict, optional
        Dictionary of plot options    

    Returns
    -------
    matplotlib.pyplot.figure or None
        A matplotlib figure unless the plot is not shown and is saved, in which case None.
    """

    options = {}
    options["declevel"] = 0
    options["eFreqI"] = 0
    options["specdir"] = projData.config.configParams["Spectra"]["specdir"]
    options["maskname"] = ""
    options["numbins"] = 40
    options["xlim"] = []
    options["maxcols"] = 4
    options["show"] = True
    options["save"] = False
    options["plotoptions"] = plotOptionsSpec()
    options = parseKeywords(options, kwargs)

    projectText(
        "Plotting histogram for statistic {}, site {} and sampling frequency {}".format(
            stat, site, sampleFreq
        )
    )

    statData = getStatisticDataForSampleFreq(
        projData,
        site,
        sampleFreq,
        stat,
        declevel=options["declevel"],
        specdir=options["specdir"],
    )
    statMeas = list(statData.keys())
    # get the statistic components
    statComponents = statData[statMeas[0]].winStats
    # get the evaluation frequency
    eFreq = statData[statMeas[0]].evalFreq[options["eFreqI"]]

    # get the mask data
    maskWindows = []
    if options["maskname"] != "":
        maskData = getMaskData(projData, site, options["maskname"], sampleFreq)
        maskWindows = maskData.getMaskWindowsFreq(
            options["declevel"], options["eFreqI"]
        )

    # plot information
    nrows, ncols = getPlotRowsAndCols(options["maxcols"], len(statComponents))
    numbins = options["numbins"]

    plotfonts = options["plotoptions"]["plotfonts"]
    fig = plt.figure(figsize=options["plotoptions"]["figsize"])
    # suptitle
    st = fig.suptitle(
        "{} histogram for {}, sampling frequency {} Hz, decimation level {} and evaluation frequency {} Hz".format(
            stat, site, sampleFreq, options["declevel"], eFreq
        ),
        fontsize=plotfonts["suptitle"],
    )
    st.set_y(0.98)

    # now plot the data
    for idx, val in enumerate(statComponents):
        ax = plt.subplot(nrows, ncols, idx + 1)
        plt.title("Histogram {}".format(val), fontsize=plotfonts["title"])

        plotData = np.empty(shape=(0))
        for meas in statMeas:
            stats = statData[meas].getStats(maskwindows=maskWindows)
            plotData = np.concatenate(
                (plotData, np.squeeze(stats[:, options["eFreqI"], idx]))
            )
        # remove infinities and nans
        plotData = plotData[np.isfinite(plotData)]

        # x axis options
        xlim = (
            kwargs["xlim"] if "xlim" in kwargs else [np.min(plotData), np.max(plotData)]
        )
        plt.xlim(xlim)
        plt.xlabel("Value", fontsize=plotfonts["axisLabel"])
        # now plot with xlim in mind
        plt.hist(plotData, numbins, range=xlim, facecolor="red", alpha=0.75)
        plt.grid()
        # y axis options
        plt.ylabel("Count", fontsize=plotfonts["axisLabel"])
        # set tick sizes
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(plotfonts["axisTicks"])

    # plot format, show and save
    fig.tight_layout(rect=[0.02, 0.02, 0.98, 0.92])     
    if options["save"]:
        impath = projData.imagePath
        sampleFreqStr = fileFormatSampleFreq(sampleFreq)
        filename = "statHist_{:s}_{:s}_{:s}_dec{:d}_efreq{:d}_{:s}".format(
            stat,
            site,
            sampleFreqStr,
            options["declevel"],
            options["eFreqI"],
            options["specdir"],
        )
        if options["maskname"] != "":
            filename = "{}_{}".format(filename, options["maskname"])
        savename = savePlot(impath, filename, fig)
        projectText("Image saved to file {}".format(savename))
    if options["show"]:
        plt.show(block=options["plotoptions"]["block"])
    if not options["show"] and options["save"]:
        plt.close(fig)
        return None        
    return fig



import os
from configobj import ConfigObj
from validate import Validator
from typing import Any

# import from package
from resistics.utilities.utilsPrint import generalPrint, errorPrint
from resistics.utilities.utilsIO import checkFilepath


def getDefaultConfigFilepath() -> str:
    """Get the default global configuration option

    Returns
    -------
    str
        Path to global config file
    """

    # use relative path from here
    path = os.path.split(__file__)[0]
    globalConfigFile = os.path.join(path, "..", "resisticsConfig.ini")
    if not checkFilepath(globalConfigFile):
        errorPrint(
            "getDefaultConfig",
            "Default configuration file could not be found",
            quitRun=True,
        )

    return globalConfigFile


def loadConfig(filepath: str = "") -> ConfigObj:
    """Get configuration information

    Parameters
    ----------
    filepath : str, optional
        The path to the configuration file

    Returns
    -------
    config : ConfigObj
        ConfigObj with global configuration parameters
    """

    configFile = getDefaultConfigFilepath()
    if filepath == "" or not checkFilepath(filepath):
        config = ConfigObj(configspec=configFile)
    else:
        config = ConfigObj(filepath, configspec=configFile)
        generalPrint("loadConfig", "Loading configuration file {:s}".format(filepath))
    validator = Validator()
    result = config.validate(validator)
    if not result:
        errorPrint("loadConfigFile", "Config file validation failed", quitRun=True)
    return config


def getDefaultParameter(param: str) -> Any:
    """Get the default global configuration option

    Parameters
    ----------
    param : str
        The config parameter
    """

    config = loadConfig()
    return config[param]


def copyDefaultConfig(filepath: str, name: str = "global copy") -> None:
    """Create copy of the global configuration file 
    
    Parameters
    ----------
    filepath : str
        The path to write the copy of the config file
    """

    config = ConfigObj(configspec=getDefaultConfigFilepath())
    validator = Validator()
    config.validate(validator, copy=True)
    # change the name of the configuration
    config["name"] = name
    with open(filepath, "wb") as f:
        config.write(f)



import os
import glob
from datetime import datetime, timedelta
import numpy as np
from typing import List, Tuple

# import from package
from resistics.dataObjects.timeData import TimeData
from resistics.ioHandlers.dataReaderInternal import DataReaderInternal
from resistics.utilities.utilsClean import removeNansSingle, removeZerosSingle


class DataReaderAscii(DataReaderInternal):
    """Data reader for ascii formatted data

    The ASCII data reader reads ascii data files and internally formatted header files. No further scaling is applied to the data values in either getUnscaledSamples or getPhysicalSamples. All the data is assumed to be in the correct units.

    In fact, if the data does not have to be calibrated, the units could be anything as long as they are internally consistent.

    Methods
    -------
    setParameters()
        Set data format parameters    
    getUnscaledSamples()
        Get the raw unscaled samples from an ascii file
    getPhysicalSamples()
        Get data in field units. Note: no further scaling is applied in this function, ascii data is assumed to be in field units
    """

    def setParameters(self) -> None:
        """Set data reader parameters

        This will vary for the different data formats. By default, setup for the internal data format.
        """

        self.headerF = glob.glob(os.path.join(self.dataPath, "*.hdr"))
        self.dataF = glob.glob(os.path.join(self.dataPath, "*.ascii"))
        self.dataByteOffset = 0
        self.dataByteSize = 4

    def getUnscaledSamples(self, **kwargs) -> TimeData:
        """Get raw data from ascii data file

        This function simply reads the lines which match the samples to be read
        
        Parameters
        ----------
        chans : List[str], optional
            List of channels to return if not all are required
        startSample : int, optional
            First sample to return
        endSample : int, optional
            Last sample to return

        Returns
        -------
        TimeData
            Time data object
        """

        # initialise chans, startSample and endSample with the whole dataset
        options = self.parseGetDataKeywords(kwargs)
        # get samples - this is inclusive
        dSamples = options["endSample"] - options["startSample"] + 1

        # loop through chans and get data
        data = {}
        for chan in options["chans"]:
            # check to make sure channel exists
            self.checkChan(chan)
            # get data file
            dFile = os.path.join(self.dataPath, self.getChanDataFile(chan))
            # read the lines
            dataChan = np.zeros(shape=(dSamples), dtype=np.float32)
            with open(dFile) as dF:
                for il, line in enumerate(dF):
                    if il > options["endSample"]:
                        break
                    if il >= options["startSample"] and il <= options["endSample"]:
                        dIndex = il - options["startSample"]
                        dataChan[dIndex] = float(line.strip())
            # set the data
            data[chan] = dataChan

        # get data start and stop time
        startTime, stopTime = self.sample2time(
            options["startSample"], options["endSample"]
        )
        # dataset comments
        comments = []
        comments.append(
            "Unscaled data {} to {} read in from measurement {}, samples {} to {}".format(
                startTime,
                stopTime,
                self.dataPath,
                options["startSample"],
                options["endSample"],
            )
        )
        comments.append("Sampling frequency {}".format(self.getSampleFreq()))
        if len(self.comments) > 0:
            comments = self.comments + comments
        return TimeData(
            sampleFreq=self.getSampleFreq(),
            startTime=startTime,
            stopTime=stopTime,
            data=data,
            comments=comments,
        )

    def getPhysicalSamples(self, **kwargs):
        """Get ascii data scaled to physical values
        
        Warnings
        --------
        No scaling happens in getPhysicalSamples. Ascii data is assumed to be properly scaled to mV for magnetic channels and mV/km for electric channels (i.e. field units)

        Parameters
        ----------
        chans : List[str]
            List of channels to return if not all are required
        startSample : int
            First sample to return
        endSample : int
            Last sample to return
        remaverage : bool
            Remove average from the data
        remzeros : bool
            Remove zeroes from the data
        remnans: bool
            Remove NanNs from the data

        Returns
        -------
        TimeData
            Time data object
        """

        options = self.parseGetDataKeywords(kwargs)
        timeData = self.getUnscaledSamples(
            chans=options["chans"],
            startSample=options["startSample"],
            endSample=options["endSample"],
        )

        # no further scaling applied to ascii data
        for chan in options["chans"]:
            # if remove zeros - False by default
            if options["remzeros"]:
                timeData.data[chan] = removeZerosSingle(timeData.data[chan])
            # if remove nans - False by default
            if options["remnans"]:
                timeData.data[chan] = removeNansSingle(timeData.data[chan])
            # remove the average from the data - True by default
            # do this after all scaling and removing nans and zeros
            if options["remaverage"]:
                timeData.data[chan] = timeData.data[chan] - np.average(
                    timeData.data[chan]
                )

        timeData.addComment(
            "Remove zeros: {}, remove nans: {}, remove average: {}".format(
                options["remzeros"], options["remnans"], options["remaverage"]
            )
        )
        return timeData



import numpy as np
import numpy.linalg as linalg
import scipy.stats as stats

# get the print
from resistics.utilities.utilsPrint import *

# The source for these functions is Robust Statisitics, Huber, 2009
# in general, linear regression is# have observations y and predictors A
# y is multiple observations/response
# x are the independent variables and is unknown
# and y is a linear function of x => y = Ax
# y = nobs
# A = nobs * nregressors
# x = nregressors


# 2 STAGE ROBUST ESTIMATES
def mmestimateModel(A, y, **kwargs):
    # get options
    options = parseKeywords(kwargs)
    intercept = options["intercept"]
    # this uses an initial mestimate with huber to give a measure of scale
    # and then a second with bisquare or hampel weights
    if "initial" in kwargs:
        if "scale" not in kwargs["initial"]:
            kwargs["initial"]["scale"] = sampleMAD0(kwargs["initial"]["resids"])
        params, resids, scale, weights = mestimateModel(
            A, y, weights="huber", initial=kwargs["initial"], intercept=intercept
        )
        # now do another, but with a different weighting function
        kwargs["initial"]["scale"] = scale
        # kwargs["initial"]["params"] = params # put the new solution in, because simply then doing bisquare, which has zero weights, might mess things up
        # kwargs["initial"]["resids"] = resids
        params2, resids2, scale2, weights2 = mestimateModel(
            A, y, weights="bisquare", initial=kwargs["initial"], intercept=intercept
        )
    else:
        params, resids, scale, weights = mestimateModel(
            A, y, weights="huber", intercept=intercept
        )
        # now do another, but with a different weighting function
        params2, resids2, scale2, weights2 = mestimateModel(
            A, y, weights="bisquare", scale=scale, intercept=intercept
        )

    return params2, resids2, scale2, weights2


def smestimateModel(A, y, **kwargs):
    # this uses an initial mestimate with huber to give a measure of scale
    # and then a second with bisquare or hampel weights
    params, residuals, scale, weights = mestimateModel(
        A, y, weights="huber", intercept=True
    )
    # now do another, but with a different weighting function
    params2, residuals2, scale2, weights2 = mestimateModel(A, y, parameters)


# def reWLS(A, y, **kwargs):
# 	# begin with a bounded S-estimate
# 	# then compute the REWLS
# 	components = setimateModel(A, y, kwargs)

# 	# now do the REWLS iterations


# ORDINARY LEAST SQUARES
def olsModel(A, y, **kwargs):
    options = parseKeywords(kwargs)
    if options["intercept"]:
        # add a constant term for the intercept
        A = np.hstack((np.ones(shape=(A.shape[0], 1), dtype="complex"), A))
    params, squareResid, rank, s = linalg.lstsq(A, y)
    resids = y - np.dot(A, params)
    return params, resids, squareResid, rank, s


# ROBUST LEAST SQUARES
# mestimate model
# good for dependent outliers
# not robust against independent outliers (called leverage points)
def mestimateModel(A, y, **kwargs):
    # get options
    options = parseKeywords(kwargs)

    # calculate the leverage
    n = A.shape[0]
    p = A.shape[1]
    pnRatio = 1.0 * p / n

    # calculate the projection matrix
    q, r = linalg.qr(A)
    Pdiag = np.empty(shape=(n), dtype="float")
    for i in range(0, n):
        Pdiag[i] = np.absolute(np.sum(q[i, :] * np.conjugate(q[i, :]))).real
    del q, r
    Pdiag = Pdiag / np.max(Pdiag)
    leverageScale = sampleMAD0(Pdiag)
    leverageWeights = getRobustLocationWeights(
        Pdiag / leverageScale, "huber"
    )  # this should nowhere be equal to zero because of the previous line

    if options["intercept"] == True:
        # add column of ones for constant term
        A = np.hstack((np.ones(shape=(A.shape[0], 1), dtype="complex"), A))

    # see whether to do an initial OLS model or whether one is provided
    if options["initial"]:
        params, resids, scale = initialFromDict(options["initial"])
    else:
        params, resids, squareResid, rank, s = olsModel(A, y)
        scale = sampleMAD0(resids)

    # if an initial model was not provided but an initial scale was, replace the one here
    if options["scale"]:
        scale = options["scale"]

    # standardised residuals and weights
    weights = (
        getRobustLocationWeights(resids / scale, options["weights"]) * leverageWeights
    )

    # iteratively weighted least squares
    iteration = 0
    while iteration < options["maxiter"]:
        # do the weighted least-squares
        Anew, ynew = weightLS(A, y, weights)
        paramsNew, squareResidNew, rankNew, sNew = linalg.lstsq(Anew, ynew)
        residsNew = y - np.dot(A, paramsNew)
        # check residsNew to make sure not all zeros (i.e. will happen in undetermined or equally determined system)
        if np.sum(np.absolute(residsNew)) < eps():
            # then return everything here
            return paramsNew, residsNew, scale, weights
        scale = sampleMAD0(residsNew)
        # standardise and calculate weights
        weightsNew = (
            getRobustLocationWeights(residsNew / scale, options["weights"])
            * leverageWeights
        )
        # increment iteration and save weightsNew
        iteration = iteration + 1
        weights = weightsNew
        params = paramsNew

        # check to see whether the change is smaller than the tolerance
        # use the R method of checking change in residuals (can check change in params)
        changeResids = linalg.norm(residsNew - resids) / linalg.norm(residsNew)
        if changeResids < eps():
            # update residuals
            resids = residsNew
            break
        # update residuals
        resids = residsNew
    return params, resids, scale, weights


# this is the s-estimate
# which has a different scaling version
# def sestimateModel(A, y, **kwargs):
# 	# get options
# 	options = parseKeywords(kwargs)
# 	#generalPrint("S-Estimate", "Using weight function = {}".format(weightFnc))
# 	if options["intercept"] == True:
# 		# add column of ones for constant term
# 		A = np.hstack((np.ones(shape=(A.shape[0],1), dtype="complex"), A))

# 	# see whether to do an initial OLS model or whether one is provided
# 	if options["initial"]:
# 		params, resids, scale = initialFromDict(options["initial"])
# 	else:
# 		params, resids, squareResid, rank, s = olsModel(A, y)
# 		scale = sampleMAD0(resids)

# 	# standardised residuals and weights
# 	resids = resids/scale
# 	weights = getRobustLocationWeights(resids, options["weights"])

# 	# iteratively weighted least squares
# 	iteration = 0
# 	while iteration < options["maxiter"]:
# 		# do the weighted least-squares
# 		Anew, ynew = weightLS(A, y, weights)
# 		componentsNew, squareResidNew, rankNew, sNew = linalg.lstsq(Anew, ynew)
# 		residualsNew = y - np.dot(A, componentsNew)
# 		# now for s estimate, the new scale is calculated differently
# 		scale = sestimateScale(residualsNew)
# 		# standardised residuals
# 		residualsNew = residualsNew/scale
# 		weightsNew = getRobustLocationWeights(residualsNew, weightFnc)

# 		# increment iteration
# 		iteration = iteration + 1
# 		weights = weightsNew

# 		# check to see whether the change is smaller than the tolerance
# 		change = linalg.norm(componentsNew-components)/linalg.norm(componentsNew)
# 		changeResids = linalg.norm(residualsNew-residuals)/linalg.norm(residualsNew)
# 		if changeResids < eps():
# 			# update components
# 			components = componentsNew
# 			break
# 		# update components
# 		components = componentsNew

# 	#generalPrint("S-Estimate", "Robust regression quit after {} iterations".format(iteration))
# 	# at the end, return the components
# 	return components, weights


# a bounded influence estimator
def chatterjeeMachler(A, y, **kwargs):
    # get options
    options = parseKeywords(kwargs)
    # generalPrint("S-Estimate", "Using weight function = {}".format(weightFnc))
    if options["intercept"] == True:
        # add column of ones for constant term
        A = np.hstack((np.ones(shape=(A.shape[0], 1), dtype="complex"), A))

    # now calculate p and n
    n = A.shape[0]
    p = A.shape[1]
    pnRatio = 1.0 * p / n

    # calculate the projection matrix
    q, r = linalg.qr(A)
    Pdiag = np.empty(shape=(n), dtype="float")
    for i in range(0, n):
        Pdiag[i] = np.absolute(np.sum(q[i, :] * np.conjugate(q[i, :]))).real
    del q, r
    # and save an array for later
    Pdiag = Pdiag / np.max(Pdiag)
    weightsNom = np.power(1.0 - Pdiag, 2)

    # weights for the first iteration
    tmp = np.ones(shape=(n), dtype="float") * pnRatio
    tmp = np.maximum(Pdiag, tmp)
    weights = np.reciprocal(tmp)

    # iteratively weighted least squares
    iteration = 0
    while iteration < options["maxiter"]:
        # do the weighted least-squares
        Anew, ynew = weightLS(A, y, weights)
        paramsNew, squareResidNew, rankNew, sNew = linalg.lstsq(Anew, ynew)
        residsNew = y - np.dot(A, paramsNew)
        # check residsNew to make sure not all zeros (i.e. will happen in undetermined or equally determined system)
        if np.sum(np.absolute(residsNew)) < eps():
            # then return everything here
            return paramsNew, residsNew, weights
        residsAbs = np.absolute(residsNew)
        residsMedian = np.median(residsAbs)
        # now compute the new weights
        weightsDenom = np.maximum(
            residsAbs, np.ones(shape=(n), dtype="float") * residsMedian
        )
        weightsNew = weightsNom / weightsDenom

        # increment iteration
        iteration = iteration + 1
        weights = weightsNew
        params = paramsNew

        if iteration > 1:
            # check to see whether the change is smaller than the tolerance
            changeResids = linalg.norm(residsNew - resids) / linalg.norm(residsNew)
            if changeResids < eps():
                # update resids
                resids = residsNew
                break
        # update resids
        resids = residsNew
    return params, resids, weights


def chatterjeeMachlerMod(A, y, **kwargs):
    # using the weights in chaterjeeMachler means that min resids val in median(resids)
    # instead, use M estimate weights with a modified residual which includes a measure of leverage
    # for this, use residuals / (1-p)^2
    # I wonder if this will have a divide by zero bug

    # now calculate p and n
    n = A.shape[0]
    p = A.shape[1]
    pnRatio = 1.0 * p / n

    # calculate the projection matrix
    q, r = linalg.qr(A)
    Pdiag = np.empty(shape=(n), dtype="float")
    for i in range(0, n):
        Pdiag[i] = np.absolute(np.sum(q[i, :] * np.conjugate(q[i, :]))).real
    del q, r
    Pdiag = Pdiag / (np.max(Pdiag) + 0.0000000001)
    locP = np.median(Pdiag)
    scaleP = sampleMAD(Pdiag)
    # bound = locP + 6*scaleP
    bound = locP + 6 * scaleP
    indices = np.where(Pdiag > bound)
    Pdiag[indices] = 0.99999
    leverageMeas = np.power(1.0 - Pdiag, 2)

    # weights for the first iteration
    # this is purely based on the leverage
    tmp = np.ones(shape=(n), dtype="float") * pnRatio
    tmp = np.maximum(Pdiag, tmp)
    weights = np.reciprocal(tmp)

    # get options
    options = parseKeywords(kwargs)
    # generalPrint("S-Estimate", "Using weight function = {}".format(weightFnc))
    if options["intercept"] == True:
        # add column of ones for constant term
        A = np.hstack((np.ones(shape=(A.shape[0], 1), dtype="complex"), A))

    # iteratively weighted least squares
    iteration = 0
    while iteration < options["maxiter"]:
        # do the weighted least-squares
        Anew, ynew = weightLS(A, y, weights)
        paramsNew, squareResidNew, rankNew, sNew = linalg.lstsq(Anew, ynew)
        residsNew = y - np.dot(A, paramsNew)
        # check residsNew to make sure not all zeros (i.e. will happen in undetermined or equally determined system)
        if np.sum(np.absolute(residsNew)) < eps():
            # then return everything here
            return paramsNew, residsNew, weights
        residsNew = residsNew / leverageMeas
        scale = sampleMAD0(residsNew)

        # standardise and calculate weights
        residsNew = residsNew / scale
        weightsNew = getRobustLocationWeights(residsNew, "huber")
        # increment iteration
        iteration = iteration + 1
        weights = weightsNew
        params = paramsNew

        if iteration > 1:
            # check to see whether the change is smaller than the tolerance
            changeResids = linalg.norm(residsNew - resids) / linalg.norm(residsNew)
            if changeResids < eps():
                # update resids
                resids = residsNew
                break
        # update resids
        resids = residsNew

    # now do the same again, but with a different function
    # do the least squares solution
    params, resids, squareResid, rank, s = olsModel(A, y)
    resids = resids / leverageMeas
    resids = resids / scale
    weights = getRobustLocationWeights(resids, "trimmedMean")
    # iteratively weighted least squares
    iteration = 0
    while iteration < options["maxiter"]:
        # do the weighted least-squares
        Anew, ynew = weightLS(A, y, weights)
        paramsNew, squareResidNew, rankNew, sNew = linalg.lstsq(Anew, ynew)
        residsNew = y - np.dot(A, paramsNew)
        # check residsNew to make sure not all zeros (i.e. will happen in undetermined or equally determined system)
        if np.sum(np.absolute(residsNew)) < eps():
            # then return everything here
            return paramsNew, residsNew, weights

        residsNew = residsNew / leverageMeas
        scale = sampleMAD0(residsNew)

        # standardise and calculate weights
        residsNew = residsNew / scale
        weightsNew = getRobustLocationWeights(residsNew, options["weights"])
        # increment iteration
        iteration = iteration + 1
        weights = weightsNew
        params = paramsNew

        # check to see whether the change is smaller than the tolerance
        changeResids = linalg.norm(residsNew - resids) / linalg.norm(residsNew)
        if changeResids < eps():
            # update resids
            resids = residsNew
            break
        # update resids
        resids = residsNew

    # at the end, return the components
    return params, resids, weights


# Another regression method based on Hadi distances
# implemented from the paper A Re-Weighted Least Squares Method for Robust Regression Estimation
# Billor, Hadi
def chatterjeeMachlerHadi(X, y, **kwargs):
    # basic info
    options = parseKeywords(kwargs)

    # for the distances, will use absX - do this before adding intercept term
    # a column of all ones will cause problems with non full rank covariance matrices
    absX = np.absolute(X)

    # now calculate p and n
    n = absX.shape[0]
    p = absX.shape[1]

    # we treat the X matrix as a multivariate matrix with n observations and p variables
    # first need to find a basic subset free of outliers
    correctionFactor = 1 + (1.0 * (p + 1) / (n - p)) + (2.0 / (n - 1 - 3 * p))
    chi = stats.chi2(p, 0)
    alpha = 0.05
    chi2bound = correctionFactor * chi.pdf(alpha / n)
    # calculate h, this is the size of the firt basic subset
    # note that this is the value h, the index of the hth element is h-1
    h = int(1.0 * (n + p + 1) / 2)  # here, only want the integer part of this
    # need to get the coordinatewise medians - this is the median of the columns
    medians = np.median(absX)
    # now compute the matrix to help calculate the distance
    A = np.zeros(shape=(p, p))
    for i in range(0, n):
        tmp = absX[i, :] - medians
        A += np.outer(tmp, tmp)
    A = 1.0 / (n - 1) * A

    # now calculate initial distances
    dInit = calculateDistCMH(n, absX, medians, A)

    # now get the h smallest values of d
    sortOrder = np.argsort(dInit)
    indices = sortOrder[0:h]
    means = np.average(absX[indices, :])
    covariance = np.cov(
        absX[indices], rowvar=False
    )  # observations in rows, columns are variables
    dH = calculateDistCMH(n, absX, means, covariance)

    # rearrange into n observations into order and partition into two initial subsets
    # one subset p+1, the n-p-1
    sortOrder = np.argsort(dH)
    indicesBasic = sortOrder[: p + 1]
    # there is a rank issue here, but ignore for now - natural observations will presumably be full rank
    means = np.average(absX[indicesBasic, :])
    covariance = np.cov(absX[indicesBasic], rowvar=False)
    dist = calculateDistCMH(n, absX, means, covariance)

    # create the basic subset
    r = p + 2
    increment = (h - r) / 100
    if increment < 1:
        increment = 1  # here, limiting to 100 iterations of this
    while r <= h:
        sortOrder = np.argsort(dist)
        indices = sortOrder[:r]  # indices start from zero, hence the - 1
        means = np.average(absX[indices])
        covariance = np.cov(absX[indices], rowvar=False)
        dist = calculateDistCMH(n, absX, means, covariance)
        if h - r > 0 and h - r < increment:
            r = h
        else:
            r += increment

    # now the second part = add more points and exclude outliers to basic set
    # all distances above r+1 = outliers
    # r = p + 1
    # increment = (n - 1 - r)/100
    while r < n:
        sortOrder = np.argsort(dist)
        dist2 = np.power(dist, 2)
        if dist2[sortOrder[r]] > chi2bound:
            break  # then leave, everything else is an outlier - it would be good if this could be saved somehow
        # otherwise, continue adding points
        sortOrder = np.argsort(dist)
        indices = sortOrder[:r]
        means = np.average(absX[indices])
        covariance = np.cov(absX[indices], rowvar=False)
        dist = calculateDistCMH(n, absX, means, covariance)
        if n - 1 - r > 0 and n - 1 - r < increment:
            r = n - 1
        else:
            r += increment

    # now with the Hadi distances calculated, can proceed to do the robust regression
    # normalise and manipulate Hadi distances
    dist = dist / np.max(dist)
    # for the median, use the basic subset
    # indicesBasic = sortOrder[:r]
    # distMedian = np.median(dist[indicesBasic]) # I am using on indicesBasic
    distMedian = np.median(dist)  # the paper suggests using the median of the complete
    tmp = np.maximum(dist, np.ones(shape=(n)) * distMedian)
    dist = np.reciprocal(tmp)
    dist2 = np.power(dist, 2)
    dist = dist2 / np.sum(dist2)

    # calculate first set of weights - this is simply dist
    weights = dist

    # now add the additional constant intercept column if required
    if options["intercept"] == True:
        # add column of ones for constant term
        X = np.hstack((np.ones(shape=(X.shape[0], 1), dtype="complex"), X))

    n = X.shape[0]
    p = X.shape[1]

    # iteratively weighted least squares
    iteration = 0
    while iteration < options["maxiter"]:
        # do the weighted least-squares
        Anew, ynew = weightLS(X, y, weights)
        paramsNew, squareResidNew, rankNew, sNew = linalg.lstsq(Anew, ynew)
        residsNew = y - np.dot(X, paramsNew)
        # check residsNew to make sure not all zeros (i.e. will happen in undetermined or equally determined system)
        if np.sum(np.absolute(residsNew)) < eps():
            # then return everything here
            return paramsNew, residsNew, weights

        residsAbs = np.absolute(residsNew)
        residsSquare = np.power(residsAbs, 2)
        residsNew = residsSquare / np.sum(residsSquare)
        residsMedian = np.median(residsAbs)

        # calculate the new weights
        tmpDenom = np.maximum(
            residsNew, np.ones(shape=(n), dtype="float") * residsMedian
        )
        tmp = (1 - dist) / tmpDenom
        weightsNew = np.power(tmp, 2) / np.sum(np.power(tmp, 2))

        # increment iteration
        iteration = iteration + 1
        weights = weightsNew
        params = paramsNew

        if iteration > 1:
            # check to see whether the change is smaller than the tolerance
            changeResids = linalg.norm(residsNew - resids) / linalg.norm(residsNew)
            if changeResids < eps():
                # update resids
                resids = residsNew
                break
        # update resids
        resids = residsNew

    # at the end, return the components
    return params, resids, weights


def calculateDistCMH(n, x, mean, covariance):
    inv = np.linalg.inv(covariance)
    dist = np.empty(shape=(n), dtype="float")
    for i in range(0, n):
        tmp = x[i, :] - mean
        dist[i] = np.sqrt(np.dot(tmp, np.dot(inv, tmp)))
    return dist


# a bounded influence estimator
# this is good against leverage points
# def schweppeModel(A, y, **kwargs):


# helper functions
# Weighted least squares:
# sqrt(weights) * y = sqrt(weights) * A * x
# is equivalent to
# A' *weights* y = A' * weights * A x (note A' transpose is the hermitian transpose)
# i.e. both y and A are multiplied the square root of the weights
def weightLS(A, y, weights):
    ynew = np.sqrt(weights) * y
    Anew = np.empty(shape=A.shape, dtype="complex")
    for col in range(0, A.shape[1]):
        Anew[:, col] = np.sqrt(weights) * A[:, col]
    return Anew, ynew


# calculate various values for returning
# from the robust regression solution
# this includes residuals and variances for
# error calculations
# def varsForReturn(residuals):


def hermitianTranspose(mat):
    return np.conjugate(np.transpose(mat))


# simple function that returns the intial model
# from a provided initial model dictionary
# this helps for two stage robust regression methods
def initialFromDict(initDict):
    return initDict["params"], initDict["resids"], initDict["scale"]


def parseKeywords(keywords):
    # general function to parse keywords
    # can be used in a number of different algorithms
    # and there will simply be redundant keywords
    outDict = defaultDictionary()
    if not keywords:
        return outDict

    if "weights" in keywords:
        outDict["weights"] = keywords["weights"]
    if "maxiter" in keywords:
        outDict["maxiter"] = keywords["maxiter"]
    if "initial" in keywords:
        outDict["initial"] = keywords["initial"]
    if "scale" in keywords:
        outDict["scale"] = keywords["scale"]
    if "intercept" in keywords:
        outDict["intercept"] = keywords["intercept"]
    return outDict


def defaultDictionary():
    outDict = {}
    outDict["weights"] = "bisquare"
    outDict["maxiter"] = maxIter()
    outDict["initial"] = False
    outDict["scale"] = False
    outDict["intercept"] = False
    return outDict


######################
### LOCATION WEIGHTING FUNCTIONS
### Functions for re-weighting least squares rows
######################
def getRobustLocationWeights(r, weight):
    # the second argument, k, is a tuning constant
    if weight == "huber":
        k = 1.345
        # k = 0.5
        return huberLocationWeights(r, k)
    elif weight == "hampel":
        k = 8
        return hampelLocationWeights(r, k)
    elif weight == "trimmedMean":
        k = 2
        return trimmedMeanLocationWeights(r, k)
    elif weight == "andrewsWave":
        k = 1.339
        return andrewsWaveLocationWeights(r, k)
    elif weight == "leastsq":
        return leastSquaresLocationWeights(r)
    else:
        # use bisquare weights
        k = 4.685
        # k = 1.0
        return bisquareLocationWeights(r, k)


# relying on numpy doing the right thing
# when dividing by zero
def huberLocationWeights(r, k):
    weights = np.ones(shape=r.size, dtype="complex")
    for idx, val in enumerate(np.absolute(r)):
        if val > k:
            weights[idx] = k / val
    return weights.real


def bisquareLocationWeights(r, k):  # biweight
    ones = np.ones(shape=(r.size), dtype="complex")
    threshR = np.minimum(ones, np.absolute(r / k))
    # threshR = np.maximum(-1*ones, threshR)
    return np.power((1 - np.power(threshR, 2)), 2).real
    # for scale weights


def hampelLocationWeights(r, k):
    a = k / 4
    b = k / 2
    weights = np.ones(shape=r.size, dtype="complex")
    for idx, val in enumerate(np.absolute(r)):
        if val > a and val <= b:
            weights[idx] = a / val
        if val > b and val <= k:
            weights[idx] = a * (k - val) / (val * (k - b))
        if val > k:
            weights[idx] = 0
    return weights.real


def trimmedMeanLocationWeights(r, k):
    weights = np.zeros(shape=r.size, dtype="complex")
    indices = np.where(np.absolute(r) <= k)
    weights[indices] = 1
    return weights.real


def andrewsWaveLocationWeights(r, k):
    weights = np.zeros(shape=r.size, dtype="complex")
    testVal = k * np.pi
    for idx, val in enumerate(np.absolute(r)):
        if val < testVal:
            weights[idx] = np.sin(val / k) / (val / k)
    return weights.real


# least squares has no weighting
def leastSquaresLocationWeights(r):
    return np.ones(shape=(r.size), dtype="complex")


######################
### SCALE WEIGHTING FUNCTIONS
### Functions for re-weighting least squares rows
######################
def getRobustScaleWeights(r, weight):
    # k is a tuning parameter
    # k = 1.56
    k = 4.685
    return bisquareScaleWeights(r, k)


def bisquareScaleWeights(r, k):
    # r = r/k
    tmp1 = 3 - 3 * np.power(r, 2) + np.power(r, 4)
    tmp2 = np.reciprocal(np.power(r, 2))
    return np.minimum(tmp1, tmp2)


# def sestimateScale(r, k):

# def alphaTrimmedScale(r, k):


# LOCATION ESTIMATORS
# The mean is not a robust estimator of location
# These are other methods of location estimation
# SCALE ESTIMATORS
# equivalent, the sd is not a robust measurement of
# dispersion
def sampleMedian(data):
    return np.median(data)


def sampleMAD(data):
    # the MAD is the median
    absData = np.absolute(data)
    mad = sampleMedian(np.absolute(absData - sampleMedian(absData)))
    return mad / 0.67448975019608171


# this uses an estimate of the location as 0
def sampleMAD0(data):
    # the MAD is the median - this should be over non zero data
    absData = np.absolute(data)
    inputIndices = np.where(absData != 0.0)
    mad = sampleMedian(absData[inputIndices])
    # mad = sampleMedian(np.absolute(data))
    return mad / 0.67448975019608171


# compute the m-estimate of location and scale
# through iteration, beginning with sample median for mean
# and madn for dispersion
def mestimate(data, **kwargs):
    location = "bisquare"
    scale = "bisquare"
    if "location" in kwargs:
        location = kwargs["location"]
    if "scale" in kwargs:
        scale = kwards["scale"]
    mean = sampleMedian(data)
    sigma = sampleMAD(data)

    iteration = 0
    n = data.size
    while iteration < maxIter():
        # calculate outlyingness
        r = (data - mean) / sigma
        # calculate new set of weights using window function
        weights1 = getRobustLocationWeights(r, location)
        weights2 = getRobustScaleWeights(r, scale)
        # now weight the data (observations)
        # in calculation of new mean and sigma
        new_mean = np.sum(weights1 * data) / np.sum(weights1)
        new_sigma2 = sigma * sigma * np.sum(weights2 * np.power(r, 2)) / (n * delta())
        new_sigma = np.sqrt(new_sigma2)

        if new_mean - mean < eps() * sigma:
            break

        # if not breaking, update mean and sigma
        mean = new_mean
        sigma = new_sigma
    # return mean and sigma
    return mean, sigma


# SIGMA FUNCTIONS
# Robust measures of outlyingness
def threeSigmaMAD(data):
    # calculate deviation from MAD
    return (data - sampleMedian(data)) / sampleMAD(data)


# ROBUST CORRELATION
# For measuring similarity between datasets


# A FEW USEFUL NUMBERS
def eps():
    # A small number of stopping iterations
    return 0.0001


def delta():
    return 0.5


def maxIter():
    return 100



import numpy as np
import scipy.interpolate as interp
from datetime import datetime, timedelta
from typing import Dict

# import from package
from resistics.dataObjects.timeData import TimeData
from resistics.utilities.utilsMath import intdiv
from resistics.utilities.utilsPrint import errorPrint


def interpolateToSecond(timeData: TimeData, inplace: bool = True) -> TimeData:
    """Interpolate data to be on the second

    Some formats of time data (e.g. SPAM) do not start on the second with their sampling. This method interpolates so that sampling starts on the second and improves interoperability with other recording formats. 

    Parameters
    ----------
    timeData : TimeData
        Time data to interpolate onto the second
    inplace :  bool, optional
        Whether to do the interpolation inplace or not. Default is True.
    
    Returns
    -------
    TimeData
        Time data interpolated to start on the second
    """

    startTimeInterp, numSamplesInterp, dataInterp = interpolateToSecondData(
        timeData.data, timeData.sampleFreq, timeData.startTime
    )
    if not inplace:
        timeData = timeData.copy()
    timeData.numSamples = numSamplesInterp
    timeData.startTime = startTimeInterp
    # calculate end timeEnd
    timeData.stopTime = timeData.startTime + timedelta(
        seconds=(1.0 / timeData.sampleFreq) * (timeData.numSamples - 1)
    )
    timeData.data = dataInterp
    timeData.addComment(
        "Time data interpolated to nearest second. New start time {}, new end time {}, new number of samples {} ".format(
            timeData.startTime, timeData.stopTime, timeData.numSamples
        )
    )
    return timeData


def interpolateToSecondData(
    data: Dict[str, np.ndarray], sampleFreq: float, startTime: datetime
) -> Dict[str, np.ndarray]:
    """Interpolate data to be on the second

    Interpolates the sampling so that it coincides with full seconds. The function also shifts the start point to the next full second
    WARNING: Do not use this method on data recording with a sampling frequency of less than 1Hz
    
    Parameters
    ----------
    data : Dict
        Dictionary with channel as keys and data as values
    sampleFreq : float
        Sampling frequency of the data
    startTime : datetime
        Time of first sample
    
    Returns
    -------
    data : Dict
        Dictionary with channel as keys and data as values

    Notes
    -----
    This function will truncate the data to the next second.

    todo:
    This function needs to be more robust for low (< 1Hz) sample frequencies as the use of microseconds and seconds makes no sense for this    
    """

    # data properties
    chans = list(data.keys())
    samplePeriod = 1.0 / sampleFreq
    # set initial vals
    numSamples = data[chans[0]].size

    # now caluclate the interpolation
    microseconds = startTime.time().microsecond
    # check if the dataset already begins on a second
    if microseconds == 0:
        return startTime, numSamples, data  # do nothing, already on the second
    # now turn microseconds into a decimal
    microseconds = microseconds / 1000000.0
    # now calculate the number of complete samples till the next second
    eps = 0.000000001
    test = microseconds
    samplesToDrop = 0
    # this loop will always either calculate till the full second or the next sample passed the full second
    while test < 1.0 - eps:
        test += samplePeriod
        samplesToDrop += 1

    # if this is exact, i.e. integer number of samples to next second, just need to drop samples
    multiple = (1.0 - microseconds) / samplePeriod
    if np.absolute(multiple - samplesToDrop) < eps:  # floating point arithmetic
        dataInterp = {}  # create a new dictionary for data
        for chan in chans:
            dataInterp[chan] = data[chan][samplesToDrop:]
        # update the other data
        numSamplesInterp = numSamples - samplesToDrop
        startTimeInterp = startTime + timedelta(
            seconds=1.0 * samplesToDrop / sampleFreq
        )
        return startTimeInterp, numSamplesInterp, dataInterp

    # if here, then we have calculated one extra for samplesToDrop
    samplesToDrop -= 1

    # now the number of samples to the next full second is not an integer
    # interpolation will have to be performed
    shift = (multiple - samplesToDrop) * samplePeriod
    sampleShift = shift / samplePeriod
    x = np.arange(0, numSamples)
    xInterp = np.arange(samplesToDrop, numSamples - 1) + sampleShift
    # calculate return vars
    numSamplesInterp = xInterp.size
    startTimeInterp = (
        startTime
        + timedelta(seconds=1.0 * samplesToDrop / sampleFreq)
        + timedelta(seconds=shift)
    )

    # do the interpolation
    dataInterp = {}
    for chan in chans:
        # interpFunc = interp.InterpolatedUnivariateSpline(x, data[chan])
        # dataInterp[chan] = interpFunc(xInterp)
        tck = interp.splrep(x, data[chan], s=0)
        dataInterp[chan] = interp.splev(xInterp, tck, der=0)

    # need to calculate how much the
    return startTimeInterp, numSamplesInterp, dataInterp


def fillGap(timeData1, timeData2):
    """Fill gap between time series
    
    Fill gaps between two different recordings. The intent is to fill the gap when recording has been interrupted and there are two data files. Both times series must have the same sampling frequency.

    Parameters
    ----------
    timeDat1 : TimeData
        Time series data
    timeData2 : TimeData
        Time series data

    Returns
    -------
    TimeData
        Time series data with gap filled
    """

    if timeData1.sampleFreq != timeData2.sampleFreq:
        errorPrint(
            "fillGap",
            "fillGap requires both timeData objects to have the same sample rate",
            quitRun=True,
        )
        return False
    sampleFreq = timeData1.sampleFreq
    sampleRate = 1.0 / sampleFreq
    timeDataFirst = timeData1
    timeDataSecond = timeData2
    if timeData1.startTime > timeData2.stopTime:
        timeDataFirst = timeData2
        timeDataSecond = timeData1
    # now want to do a simple interpolation between timeDataFirst and timeDataSecond
    # recall, these times are inclusive, so want to do the samples in between
    # this is mostly for clarity of programming
    gapStart = timeDataFirst.stopTime + timedelta(seconds=sampleRate)
    gapEnd = timeDataSecond.startTime - timedelta(seconds=sampleRate)
    # calculate number of samples in the gap
    numSamplesGap = (
        int(round((gapEnd - gapStart).total_seconds() * sampleFreq)) + 1
    )  # add 1 because inclusive
    # now want to interpolate
    newData = {}
    for chan in timeDataFirst.chans:
        startVal = timeDataFirst.data[chan][-1]
        endVal = timeDataSecond.data[chan][0]
        increment = 1.0 * (endVal - startVal) / (numSamplesGap + 2)
        fillData = np.zeros(shape=(numSamplesGap), dtype=timeDataFirst.data[chan].dtype)
        for i in range(0, numSamplesGap):
            fillData[i] = startVal + (i + 1) * increment
        newData[chan] = np.concatenate(
            [timeDataFirst.data[chan], fillData, timeDataSecond.data[chan]]
        )
    # return a new time data object
    # deal with the comment
    comment = (
        ["-----------------------------", "TimeData1 comments"]
        + timeDataFirst.comments
        + ["-----------------------------", "TimeData2 comments"]
        + timeDataSecond.comments
    )
    comment += ["-----------------------------"] + [
        "Gap filled from {} to {}".format(gapStart, gapEnd)
    ]
    return TimeData(
        sampleFreq=sampleFreq,
        startTime=timeDataFirst.startTime,
        stopTime=timeDataSecond.stopTime,
        data=newData,
        comments=comment,
    )
